from spreco.model import nn
from spreco.common.custom_adam import AdamOptimizer
from spreco.common.options import DATA_CHNS

import numpy as np
from tf_slim import arg_scope
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

class pixelcnn():
    
    def __init__(self, config):
        self.config         = config
        self.forward        = tf.make_template('forward', self.body)

    def init_training(self):
        """
        initialize placeholder for training
        """
        self.learning_rate = tf.placeholder(tf.float32, shape=[])
        self.x  = [tf.placeholder(tf.float32, 
                               shape=[self.config['batch_size']]+self.config['input_shape'], name="input_%d"%i
                               ) for i in range(self.config['nr_gpu'])]
        if self.config['conditional']:
            self.t = [tf.placeholder(tf.float32, shape=[self.config['batch_size']]) for _ in range(self.config['nr_gpu'])]
        else:
            self.t = [None for _ in range(self.config['nr_gpu'])]

        if self.config['conditional']:
            self.ins_outs     = {'inputs': self.x, 't': self.t}
        else:
            self.ins_outs     = {'inputs': self.x}


    def init_inference(self, batch_size=None):
        """
        initialize placeholder for inference, on a single gpu

        """

        if batch_size is None:
            batch_size = self.config['batch_size']
        

        self.x = tf.placeholder(tf.float32, shape=[batch_size]+self.config['input_shape'], name='input_0')

        if self.config['conditional']:
            self.t = tf.placeholder(tf.float32, shape=[batch_size], name='input_1')
        else:
            self.t = None

    def prep(self, export=False):

        model_opt = {'nr_resnet': self.config['nr_resnet'],
                     'nr_filters': self.config['nr_filters'],
                     'nr_logistic_mix': self.config['nr_logistic_mix'],
                     'data_chns': self.config['data_chns'],
                     'rlt': self.config['rlt'],
                     'layer_norm': self.config['layer_norm']
                    }

        if self.config['data_chns'] == DATA_CHNS.CPLX:
            if self.config['rlt'] == 1:
                # assume the linear dependence between real and imaginary parts
                from spreco.model.logistic_loss import discretized_mix_logistic_loss_2 as loss_func
            elif self.config['rlt'] == 2:
                # assume independence between real and imaginary parts
                from spreco.model.logistic_loss import discretized_mix_logistic_loss_idp_2 as loss_func
            elif self.config['rlt'] == 3:
                # assume joint distribution
                from spreco.model.logistic_loss import discretized_mix_logistic_loss_bivariate as loss_func
            else:
                raise Exception("please check the assumption of relationship between real and imaginary")

        if self.config['data_chns'] == DATA_CHNS.MAG:
            from spreco.model.logistic_loss import discretized_mix_logistic_loss_1 as loss_func

        if not export:
            self.init_training()
            init_pass = self.forward(self.x[0], init=True, dropout_p=self.config['dropout_rate'], **model_opt) # initialization of parameters
            all_params = tf.trainable_variables()

            loss      = []
            grads     = []
            loss_test = []
            optimizer = AdamOptimizer(self.learning_rate, beta1=0.9, beta2=0.999)

            for i in range(self.config['nr_gpu']):
                with tf.device('/gpu:%d'%i):
                    # train
                    logits = self.forward(self.x[i], ema=None, dropout_p=self.config['dropout_rate'], **model_opt)
                    loss.append(loss_func(tf.stop_gradient(self.x[i]), logits, itg_interval=self.config['itg_interval']))
                    
                    gvs = optimizer.compute_gradients(loss[-1], all_params)
                    gvs = [(k, v) for (k, v) in gvs if k is not None]
                    grads.append(gvs)

                    # test
                    logits = self.forward(self.x[i], ema=None, dropout_p=0.0, **model_opt)
                    loss_test.append(loss_func(self.x[i], logits, itg_interval=self.config['itg_interval']))

            with tf.device('/gpu:0'):
                for i in range(1,self.config['nr_gpu']):
                    loss[0]      += loss[i]
                    loss_test[0] += loss_test[i]

            grads_avg = optimizer.average_gradients(grads)

            self.train_op    = optimizer.apply_gradients(grads_avg)
            self.loss_train  = loss[0]/(np.log(2.0)*np.prod(self.config['input_shape'])*self.config['batch_size']*self.config['nr_gpu'])
            self.loss_test   = loss_test[0]/(np.log(2.0)*np.prod(self.config['input_shape'])*self.config['batch_size']*self.config['nr_gpu'])
        else:
            self.init_inference(1)
            init_pass = self.forward(self.x, init=True, dropout_p=0., **model_opt) 
            self.out = self.forward(self.x, dropout_p=0., **model_opt)

            self.loss = loss_func(self.x, self.out)/(np.log(2.0)*np.prod(self.config['input_shape'])*1)

            output = tf.identity(tf.stack([self.loss, tf.zeros_like(self.loss)], axis=-1), name='output_0')
            self.grads = tf.squeeze(tf.gradients(self.loss, self.x), name='grad_0')
            grad_ys = tf.placeholder(tf.float32, shape=[2], name='grad_ys_0')


    @staticmethod
    def body(x, h=None, init=False, ema=None, layer_norm=False, dropout_p=0.5, nr_resnet=3, nr_filters=160, nr_logistic_mix=10, resnet_nonlinearity='concat_elu', data_chns=DATA_CHNS.CPLX, rlt=1):
        """
        We receive a Tensor x of shape (N,H,W,D1) (e.g. (12,32,32,3)) and produce
        a Tensor x_out of shape (N,H,W,D2) (e.g. (12,32,32,100)), where each fiber
        of the x_out tensor describes the predictive distribution for the RGB at
        that position.
        'h' is an optional N x K matrix of values to condition our generative model on
        """

        if data_chns==DATA_CHNS.MAG:
            x = 2*x - 1   #TODO should be moved to data pipe

        counters = {}
        with arg_scope([nn.conv2d, nn.deconv2d, nn.gated_resnet, nn.dense, nn.layer_norm], counters=counters, init=init, ema=ema, dropout_p=dropout_p):

            # parse resnet nonlinearity argument 
            # TODO should be moved to model/utils.py
            if resnet_nonlinearity == 'concat_elu':
                resnet_nonlinearity = nn.concat_elu
            elif resnet_nonlinearity == 'elu':
                resnet_nonlinearity = tf.nn.elu
            elif resnet_nonlinearity == 'relu':
                resnet_nonlinearity = tf.nn.relu
            else:
                raise('resnet nonlinearity ' + resnet_nonlinearity + ' is not supported')

            with arg_scope([nn.gated_resnet], nonlinearity=resnet_nonlinearity, h=h):

                # ////////// up pass through pixelCNN ////////
                xs = nn.int_shape(x)
                x_pad = tf.concat([x,tf.ones(xs[:-1]+[1])],3) # add channel of ones to distinguish image from padding later on
                u_list = [nn.down_shift(nn.down_shifted_conv2d(x_pad, num_filters=nr_filters, filter_size=[2, 3]))] # stream for pixels above
                ul_list = [nn.down_shift(nn.down_shifted_conv2d(x_pad, num_filters=nr_filters, filter_size=[1,3])) + \
                            nn.right_shift(nn.down_right_shifted_conv2d(x_pad, num_filters=nr_filters, filter_size=[2,1]))] # stream for up and to the left
                
                # 1
                if layer_norm:
                    u_list[-1] = nn.layer_norm(u_list[-1])
                    ul_list[-1] = nn.layer_norm(ul_list[-1])

                for rep in range(nr_resnet):
                
                    u_list.append(nn.gated_resnet(u_list[-1], conv=nn.down_shifted_conv2d,))
                    ul_list.append(nn.gated_resnet(ul_list[-1], u_list[-1], conv=nn.down_right_shifted_conv2d))

                u_list.append(nn.down_shifted_conv2d(u_list[-1], num_filters=nr_filters, stride=[2, 2]))
                ul_list.append(nn.down_right_shifted_conv2d(ul_list[-1], num_filters=nr_filters, stride=[2, 2]))

                # 2
                if layer_norm:
                    u_list[-1] = nn.layer_norm(u_list[-1])
                    ul_list[-1] = nn.layer_norm(ul_list[-1])

                for rep in range(nr_resnet):
                
                    u_list.append(nn.gated_resnet(u_list[-1], conv=nn.down_shifted_conv2d))
                    ul_list.append(nn.gated_resnet(ul_list[-1], u_list[-1], conv=nn.down_right_shifted_conv2d))

                u_list.append(nn.down_shifted_conv2d(u_list[-1], num_filters=nr_filters, stride=[2, 2]))
                ul_list.append(nn.down_right_shifted_conv2d(ul_list[-1], num_filters=nr_filters, stride=[2, 2]))

                # 3
                if layer_norm:
                    u_list[-1] = nn.layer_norm(u_list[-1])
                    ul_list[-1] = nn.layer_norm(ul_list[-1])

                for rep in range(nr_resnet):
                
                    u_list.append(nn.gated_resnet(u_list[-1], conv=nn.down_shifted_conv2d))
                    ul_list.append(nn.gated_resnet(ul_list[-1], u_list[-1], conv=nn.down_right_shifted_conv2d))

                # remember nodes
                for t in u_list+ul_list:
                    tf.add_to_collection('checkpoints', t)

                # /////// down pass ////////
                
                u = u_list.pop()
                ul = ul_list.pop()
                
                # 1
                if layer_norm:
                    u = nn.layer_norm(u)
                    ul = nn.layer_norm(ul)
                
                for rep in range(nr_resnet):
                
                    u = nn.gated_resnet(u, u_list.pop(), conv=nn.down_shifted_conv2d)
                    ul = nn.gated_resnet(ul, tf.concat([u, ul_list.pop()],3), conv=nn.down_right_shifted_conv2d)
                    tf.add_to_collection('checkpoints', u)
                    tf.add_to_collection('checkpoints', ul)
                

                u = nn.down_shifted_deconv2d(u, num_filters=nr_filters, stride=[2, 2])
                ul = nn.down_right_shifted_deconv2d(ul, num_filters=nr_filters, stride=[2, 2])
                
                # 2
                if layer_norm:
                    u = nn.layer_norm(u)
                    ul = nn.layer_norm(ul)

                for rep in range(nr_resnet+1):
                
                    u = nn.gated_resnet(u, u_list.pop(), conv=nn.down_shifted_conv2d)
                    ul = nn.gated_resnet(ul, tf.concat([u, ul_list.pop()],3), conv=nn.down_right_shifted_conv2d)
                    tf.add_to_collection('checkpoints', u)
                    tf.add_to_collection('checkpoints', ul)            

                u = nn.down_shifted_deconv2d(u, num_filters=nr_filters, stride=[2, 2])
                ul = nn.down_right_shifted_deconv2d(ul, num_filters=nr_filters, stride=[2, 2])
                

                #  3
                if layer_norm:
                    u = nn.layer_norm(u)
                    ul = nn.layer_norm(ul)

                for rep in range(nr_resnet+1):
                
                    u = nn.gated_resnet(u, u_list.pop(), conv=nn.down_shifted_conv2d)
                    ul = nn.gated_resnet(ul, tf.concat([u, ul_list.pop()],3), conv=nn.down_right_shifted_conv2d)
                    tf.add_to_collection('checkpoints', u)
                    tf.add_to_collection('checkpoints', ul)

                """
                nr logits 3, 6 or 10 depends on the number of image channel magnitude, real/image or red/green/blue
                """
                if data_chns == DATA_CHNS.RGB:
                    nr_logits = 10
                if data_chns == DATA_CHNS.CPLX:
                    if rlt == 1:
                        nr_logits = 6
                    elif rlt == 2:
                        nr_logits = 5
                    elif rlt == 3:
                        nr_logits = 5
                    else:
                        raise Exception("rlt shoulde be 1 or 2 or 3.")
                if data_chns == DATA_CHNS.MAG:
                    nr_logits = 3
                    
                x_out = nn.nin(tf.nn.elu(ul),nr_logits*nr_logistic_mix) 

                assert len(u_list) == 0
                assert len(ul_list) == 0

                return x_out



import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

def int_shape(x):
    shape = x.get_shape()
    if None in shape:
        return shape
    else:
        return list(map(int, shape))

def get_variable(name, stop_grad=False, ema=None, **kwargs):
    """ utility to get variable"""

    if stop_grad:
        v = tf.get_variable(name, **kwargs)
        v = tf.stop_gradient(v)
    
    else:
        v = tf.get_variable(name, **kwargs)
    
    if ema is not None:
        v = ema.average(v)
        
    return v

def get_name(layer_name, counters):
    """ utility to keep track of layer names"""

    if not layer_name in counters:
        counters[layer_name] = 0

    name = layer_name + '_' + str(counters[layer_name])
    counters[layer_name] += 1
    return name

def down_shift(x):
    xs = int_shape(x)
    return tf.concat([tf.zeros([xs[0],1,xs[2],xs[3]]), x[:,:xs[1]-1,:,:]],1)

def right_shift(x):
    xs = int_shape(x)
    return tf.concat([tf.zeros([xs[0],xs[1],1,xs[3]]), x[:,:,:xs[2]-1,:]],2)

def concat_elu(x):
    """
    like concatenated ReLU (http://arxiv.org/abs/1603.05201), but then with ELU
    """
    axis = len(x.get_shape())-1
    return tf.nn.elu(tf.concat([x, -x], axis))

def get_nonlinearity(strs):

    if strs == 'relu':
        return tf.nn.relu
    elif strs == 'elu':
        return tf.nn.elu
    elif strs == 'swish':
        return tf.nn.swish
    elif strs == 'softplus':
        return tf.nn.softplus
    else:
        raise NotImplementedError('nonlinearity function does not exist!')



from spreco.model.refine_net_sde import cond_refine_net_plus
from spreco.model.pixelcnn import pixelcnn
from spreco.common.custom_adam import AdamOptimizer
from spreco.common import utils, ops


import tqdm
import tensorflow.compat.v1 as tf
import numpy as np

class sde():
    """
    class for generative modeling through stochastic diffusion equations (by Yang Song)
    only variance exploding sde is implemented hereafter.
    """

    def __init__(self, config): #sigma_min=0.01, sigma_max=50., eps=1e-5, N=1000):
        """
        eps, the smallest time step to sample from
        """
        if config['net'] == 'refine':
            self.net = cond_refine_net_plus(config)
        else:
            raise ValueError('Method %s not recognized.'%config['net'])

        self.config    = config
        self.sigma_min = config['sigma_min']
        self.sigma_max = config['sigma_max']
        self.N         = config['N']
        self.T         = 1.
        self.eps       = config['eps']
        self.sigmas    = tf.exp(tf.linspace(tf.log(self.sigma_min), tf.log(self.sigma_max), self.N))

        self.learning_rate = tf.placeholder(tf.float32, shape=[])
        ###
        self.x       = [tf.placeholder(tf.float32, 
                               shape=[self.config['batch_size']]+self.config['input_shape'], name="input_%d"%i
                               ) for i in range(self.config['nr_gpu'])]
        self.t            = [tf.placeholder(tf.float32,
                               shape=[self.config['batch_size']]) for _ in range(config['nr_gpu'])]
        self.ins_outs     = {'inputs': self.x, 't': self.t}
        self.seed         = config['seed']


    def sde(self, x, t):

        sigma     = self.sigma_min * (self.sigma_max / self.sigma_min)**t
        drift     = tf.zeros_like(x)

        diffusion = sigma * tf.sqrt(2 * (tf.log(self.sigma_max) - tf.log(self.sigma_min)))

        return drift, diffusion

    def marginal_prob(self, x, t):
        """
        sigma(t)
        """
        std = self.sigma_min * (self.sigma_max / self.sigma_min) ** t
        return x, std

    def prior_sampling(self, shape):
        """
        x(T) ~ N(0, sigma_max)
        """
        return tf.random.normal(shape, seed=self.seed) * self.sigma_max

    def discretize(self, x, t):

        timestep  = tf.cast(t * (self.N - 1) / self.T, tf.int32)
        sigma     = tf.gather(self.sigmas, timestep)
        bools     = tf.equal(timestep, tf.zeros_like(t, tf.int32))
        adj_sigma = tf.where(bools, tf.zeros_like(timestep, tf.float32), tf.gather(self.sigmas, tf.cast(timestep-1, tf.int32)))

        f         = tf.zeros_like(x)
        g         = tf.sqrt(sigma ** 2 - adj_sigma ** 2)

        return f, g

    def reverse_sde(self, x, t):
        """
        create the drift and diffusion functions for the reverse SDE/ODE
        """
        drift, diffusion = self.sde(x, t)
        score = self.score(x, t)
        drift = drift - diffusion ** 2 * score

        return drift, diffusion

    def reverse_discrete(self, x, t):
        f, G  = self.discretize(x, t)
        rev_f = f - G ** 2 * self.score(x, t)
        rev_g = G 
        return rev_f, rev_g

    def score(self, x, t, continuous = True):

        if continuous:
            labels = self.marginal_prob(tf.zeros_like(x), t)[1]
        else:
            labels = self.T - t
            labels = labels*(sde.N - 1)
            labels = np.round(labels)
        
        score = self.net.forward(x, labels)

        return score

    def loss(self, x, t, continuous=True, likelihood_weighting=False):
        """
        """
        shape = x.shape        
        z = tf.random.normal(shape, seed=self.seed)

        mean, std    = self.marginal_prob(x, t)
        std          = std[:, tf.newaxis, tf.newaxis, tf.newaxis]
        perturbed_x  = mean +  std * z
        score        = self.score(perturbed_x, t, continuous)

        if not likelihood_weighting:

            tmp       = tf.reshape(tf.math.square(score * std + z), (shape[0], -1))
            if self.config['reduce_mean']:
                loss      = tf.reduce_mean(tmp, axis=-1)
            else:
                loss      = tf.reduce_sum(tmp, axis=-1)

        else:

            g2   = self.sde(tf.zeros_like(x), t)[1] ** 2
            if self.config['reduce_mean']:
                loss = tf.reduce_mean(tf.math.square(score, z * 1. / std), axis=-1)
            else:
                loss = tf.reduce_sum(tf.math.square(score, z * 1. / std), axis=-1)
            loss = loss * g2

        loss = tf.reduce_mean(loss)

        return loss
    
    def prep(self, export =True):
        """create training objective"""

        if not export:
            _          = self.loss(self.x[0], self.t[0])
            all_params = tf.trainable_variables()

            loss      = []
            grads     = []
            loss_test = []

            optimizer = AdamOptimizer(self.learning_rate, beta1=0.9, beta2=0.999)

            for i in range(self.config['nr_gpu']):
                with tf.device('/gpu:%d'%i):
                    # train
                    loss.append(self.loss(self.x[i], self.t[i]))

                    gvs = optimizer.compute_gradients(loss[-1], all_params)
                    gvs = [(k, v) for (k, v) in gvs if k is not None]
                    grads.append(gvs)

                    # test
                    loss_test.append(self.loss(self.x[i], self.t[i]))
                
            with tf.device('/gpu:0'):
                for i in range(1, self.config['nr_gpu']):
                    loss[0] += loss[i]
                    loss_test[0] += loss_test[i]
                
            grads_avg = optimizer.average_gradients(grads)

            self.train_op = optimizer.apply_gradients(grads_avg)
            self.loss_train = loss[0]/self.config['nr_gpu']
            self.loss_test  = loss_test[0]/self.config['nr_gpu']
        else:
            _          = self.loss(self.x[0], self.t[0])
            all_params = tf.trainable_variables()


class posterior_sampler():

    def __init__(self, sde, steps, target_snr, nr_samples, burn_in, burn_t, ode=False, ext_iter=0, disable_z=False, use_pixelcnn=False, pixelcnn_reg=None):
        self.sde        = sde
        self.ode        = ode
        self.steps      = steps
        self.target_snr = target_snr
        self.nr_samples = nr_samples
        self.burn_in    = burn_in
        self.burn_t     = burn_t
        self.burn_flag  = True
        self.ext_iter   = ext_iter
        self.disable_z        = disable_z
        self.use_pixelcnn     = use_pixelcnn
        self.pixelcnn_reg     = pixelcnn_reg
        self.mask = None
        self.coilsen = None
        self.shape = None
        self.und_ksp = None

    def get_grad_logp_p(self, pixelcnn_config, pixelcnn_path):

        tf.reset_default_graph()

        ins_pixelcnn = pixelcnn(pixelcnn_config)
        ins_pixelcnn.prep(True, batch_size=self.nr_samples)

        saver     = tf.train.Saver()
        tf_config = tf.ConfigProto()
        tf_config.gpu_options.allow_growth=True
        sess    = tf.Session(config=tf_config)
        sess.run(tf.global_variables_initializer())
        saver.restore(sess, pixelcnn_path)

        def grad_logp_p(x):
            return sess.run(ins_pixelcnn.grads, {ins_pixelcnn.x: x})

        self.grad_logp_p = grad_logp_p

    def conditional_ancestral_sampler(self, x, t, sess, AHA, AHy, s_stepsize, st=100, skip=1):
      
        @tf.function
        def get_noise_1(x, t):
            shape = x.shape
            shape = [1] + shape[1:]
            
            timestep  = tf.cast(t * (self.sde.N - 1) / self.sde.T, tf.int32)
            sigma     = tf.gather(self.sde.sigmas, timestep)
            bools     = tf.equal(timestep, tf.zeros_like(t, tf.int32))
            adj_sigma = tf.where(bools, tf.zeros_like(timestep, tf.float32), tf.gather(self.sde.sigmas, tf.cast(timestep-1, tf.int32)))
            diff2 = (sigma ** 2 - adj_sigma ** 2)
            tau = tf.sqrt(adj_sigma ** 2 * diff2 / (sigma ** 2))
            diff2 = diff2[:, tf.newaxis, tf.newaxis, tf.newaxis]
            tau = tau[:, tf.newaxis, tf.newaxis, tf.newaxis]

            noise = tf.random.normal(shape, seed=self.sde.seed)*tau
            return tau, diff2, noise
        
        @tf.function
        def get_noise_2(x, t):
            shape = x.shape
            shape = [self.nr_samples] + shape[1:]
            timestep  = tf.cast(t * (self.sde.N - 1) / self.sde.T, tf.int32)
            sigma     = tf.gather(self.sde.sigmas, timestep)
            bools     = tf.equal(timestep, tf.zeros_like(t, tf.int32))
            adj_sigma = tf.where(bools, tf.zeros_like(timestep, tf.float32), tf.gather(self.sde.sigmas, tf.cast(timestep-1, tf.int32)))
            diff2 = (sigma ** 2 - adj_sigma ** 2)
            tau = tf.sqrt(adj_sigma ** 2 * diff2 / (sigma ** 2))
            diff2 = diff2[:, tf.newaxis, tf.newaxis, tf.newaxis]
            tau = tau[:, tf.newaxis, tf.newaxis, tf.newaxis]

            noise = tf.random.normal(shape, seed=self.sde.seed)*tau
            return diff2, tau, noise


        score_op = self.sde.score(x, t)
        get_noise_op_1 = get_noise_1(x, t)
        get_noise_op_2 = get_noise_2(x, t)

        if self.use_pixelcnn:
            self.get_grad_logp_p(self.pixelcnn_reg['pixelcnn_config'], self.pixelcnn_reg['pixelcnn_path'])

        x_val     = np.random.rand(*(utils.cplx2float(AHy).shape))
        t_vals    = np.linspace(self.sde.T, self.sde.eps, self.sde.N)

        xs      = []


        if not self.burn_in:
            cur_samples = self.nr_samples
            x_val = np.concatenate([x_val]*cur_samples)
        else:
            cur_samples=1

        for t_i in tqdm.tqdm(t_vals[st:-1:skip]):
            for _ in range(self.steps):

                if self.burn_in: # just one time
                    if t_i < self.burn_t and self.burn_flag:
                        print("burned")
                        cur_samples = self.nr_samples
                        x_val = np.concatenate([x_val]*cur_samples)
                        self.burn_flag = False

                if self.burn_in and t_i > self.burn_t:
                    diff2, tau, noise = sess.run(get_noise_op_1, feed_dict={x: x_val, t: [t_i]*cur_samples})
                else:
                    diff2, tau, noise = sess.run(get_noise_op_2, feed_dict={x: x_val, t: [t_i]*cur_samples})

                score =  self.target_snr*sess.run(score_op,  {x: x_val, t: [t_i]*cur_samples})
                
                float_shape = x_val.shape
                _, sigma = self.sde.marginal_prob(AHy, t_i)

                if not self.disable_z:
                    z = utils.float2cplx(sess.run(tf.random.normal(float_shape, seed=self.sde.seed)*sigma))
                    noisy_x = utils.cplx2float(AHy + np.squeeze(AHA(z)))
                else:
                    noisy_x = utils.cplx2float(AHy)

                grad_data_fidelity = AHA(utils.float2cplx(x_val))
                grad_data_fidelity = utils.cplx2float(grad_data_fidelity)
                if self.ode:
                    x_val = x_val + diff2*score - tau*s_stepsize*grad_data_fidelity + s_stepsize*tau*noisy_x
                else:
                    x_val = x_val + diff2*score - tau*s_stepsize*grad_data_fidelity + s_stepsize*tau*noisy_x + noise

                if self.use_pixelcnn:
                    scale = np.max(abs(utils.float2cplx(x_val)))
                    print(scale)
                    x_val = x_val/scale
                    select = np.random.choice(2, x_val.shape, p=[self.pixelcnn_reg['dropout'], 1 - self.pixelcnn_reg['dropout']])
                    grads = self.grad_logp_p(x_val)
                    x_val = x_val - select*grads*self.pixelcnn_reg['lamb']
                    iterkspace = ops.A_cart(utils.float2cplx(x_val), self.coilsen[np.newaxis, ...], 1-self.mask[np.newaxis, ...], self.shape, axis=(1,2))
                    x_val = utils.cplx2float(ops.AT_cart(self.und_ksp+iterkspace, self.coilsen[np.newaxis, ...], np.ones_like(self.mask[np.newaxis, ...]), self.shape, axis=(1,2)))
                    x_val = x_val*scale

                xs.append(x_val)

        if self.ext_iter != 0:
            t_i = t_vals[-2]
            for _ in range(self.ext_iter):

                diff2, _, noise = sess.run(get_noise_op_2, feed_dict={x: x_val, t: [t_i]*cur_samples})
                score =  self.target_snr*sess.run(score_op,  {x: x_val, t: [t_i]*cur_samples})
                x_ = utils.cplx2float(AHy)

                grad_data_fidelity = AHA(utils.float2cplx(x_val))
                grad_data_fidelity = utils.cplx2float(grad_data_fidelity)

                x_val = x_val + diff2*score - tau*s_stepsize*grad_data_fidelity + s_stepsize*tau*x_

                xs.append(x_val)

        return xs


from spreco.model import nn, utils

import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tf_slim import add_arg_scope
from tf_slim import arg_scope

@add_arg_scope
def cond_crp_block(x, h, nr_filters, nr_stages, nonlinearity, normalizer, **kwargs):
    """
    chained residual pool block
    """
    x = nonlinearity(x)
    path = x
    if nr_filters is None:
        nr_filters = nn.int_shape(x)[-1]
    for _ in range(nr_stages):
        path = normalizer(path, h)
        path = tf.nn.avg_pool2d(path, ksize=[5,5], strides=1, padding="SAME") #avg_pool2d
        path = nn.conv2d_plus(path, nr_filters, nonlinearity=None, bias=False, scope='cond_crp')  # don't need bias 
        x = path + x
    return x

@add_arg_scope
def cond_rcu_block(x, h, nr_filters, nr_resnet, nr_stages, nonlinearity, normalizer, **kwargs):
    """
    residual convolution unit
    """
    if nr_filters is None:
        nr_filters = nn.int_shape(x)[-1]
    for _ in range(nr_resnet):
        residual = x
        for _ in range(nr_stages):
            x = normalizer(x, h)
            x = nonlinearity(x)
            x = nn.conv2d_plus(x, nr_filters, nonlinearity=None, bias=False, scope='cond_rcu')  # don't need bias
        x += residual
    return x

@add_arg_scope
def cond_msf_block(blocks, h, nr_filters, out_shape, normalizer, **kwargs):
    """
    multi-resolution fusion
    blocks -> a list or tuple of blocks passed to msf 
    out_shape -> [batch_size, dim_0, dim_1, 2]
    """
    sums = []
    
    for i in range(len(blocks)):
        xl_out = normalizer(blocks[i], h)
        if nr_filters is None:
            nr_filters = nn.int_shape(blocks[i])[-1]
        xl_out = nn.conv2d_plus(xl_out, nr_filters, nonlinearity=None, scope='cond_msf')
        xl_out = tf.image.resize(xl_out, out_shape, method='bilinear')
        sums.append(xl_out)
    return tf.reduce_sum(sums, axis=0)


@add_arg_scope
def cond_refine_block(blocks, h, nr_filters, out_shape, nonlinearity, normalizer, end=False, **kwargs):
    """
    refine block
    """
    outs = []

    for i in range(len(blocks)):
        outs.append(cond_rcu_block(blocks[i], h, nr_filters=None, nr_resnet=2, nr_stages=2, nonlinearity=nonlinearity, normalizer=normalizer))
    
    if len(blocks) > 1:
        y = cond_msf_block(outs, h, nr_filters=nr_filters, out_shape=out_shape, normalizer=normalizer)
    else:
        y = outs[0]
    
    y = cond_crp_block(y, h, nr_filters=None, nr_stages=2, nonlinearity=nonlinearity, normalizer=normalizer)
    y = cond_rcu_block(y, h, nr_filters, nr_resnet=3 if end else 1, nr_stages=2, nonlinearity=nonlinearity, normalizer=normalizer)
    
    return y

@add_arg_scope
def cond_res_block(x, h, out_filters, nonlinearity, normalizer, rescale=False, **kwargs):
    """
    resnet block
    out_filters is output_dims/feature
    """
    in_filters = nn.int_shape(x)[-1]
    x_skip = x
    x = normalizer(x, h) 
    x = nonlinearity(x)
    if rescale:
        x = nn.conv2d_plus(x, in_filters, nonlinearity=None, scope='cond_res', **kwargs)
    else:
        x = nn.conv2d_plus(x, out_filters, nonlinearity=None, scope='cond_res', **kwargs)

    x = normalizer(x, h)
    x = nonlinearity(x)

    x = nn.conv2d_plus(x, out_filters, nonlinearity=None, scope='cond_res', **kwargs)
    if 'dilation' not in kwargs.keys() and rescale:
        x = tf.nn.avg_pool2d(x, ksize=(1,2,2,1), strides=(1,2,2,1), padding='SAME')
    
    if out_filters == in_filters and not rescale:
        shortcut = x_skip
    else:
        if 'dilation' not in kwargs.keys():
            shortcut = nn.conv2d_plus(x_skip, out_filters, filter_size=[1,1])
            shortcut = tf.nn.avg_pool2d(shortcut, ksize=(1,2,2,1), strides=(1,2,2,1), padding='SAME')
        else:
            shortcut = nn.conv2d_plus(x_skip, out_filters)

    return shortcut + x

class cond_refine_net():

    def __init__(self, config, chns=2, normalizer=nn.cond_instance_norm_plus):
        self.chns          = chns
        self.nr_filters    = config['nr_filters']
        self.nr_classes    = config['nr_levels']
        self.nonlinearity  = utils.get_nonlinearity(config['nonlinearity'])
        self.normalizer    = normalizer
        self.counters      = {}
        self.affine_x      = config['affine_x']
        self.forward       = tf.make_template('forward', self.body)

    def body(self, x, h):
        """
        multi level refine net conditional on y
        """
        if self.affine_x:
            x = 2*x - 1

        with arg_scope([nn.conv2d_plus, cond_refine_block, cond_crp_block, cond_rcu_block, cond_msf_block, cond_res_block, nn.cond_instance_norm_plus],
                                 nonlinearity=self.nonlinearity, counters=self.counters, normalizer=self.normalizer, nr_classes=self.nr_classes):

            x_level_0 = nn.conv2d_plus(x, num_filters=1*self.nr_filters, nonlinearity=None)
            x_level_1_0 = cond_res_block(x_level_0, h, out_filters=1*self.nr_filters, rescale=False)
            x_level_1_1 = cond_res_block(x_level_1_0, h, out_filters=1*self.nr_filters, rescale=False)
            x_level_2_0 = cond_res_block(x_level_1_1, h, out_filters=2*self.nr_filters, rescale=True)
            x_level_2_1 = cond_res_block(x_level_2_0, h, out_filters=2*self.nr_filters, rescale=False)
            x_level_3_0 = cond_res_block(x_level_2_1, h, out_filters=2*self.nr_filters, rescale=True, dilation=2)
            x_level_3_1 = cond_res_block(x_level_3_0, h, out_filters=2*self.nr_filters, rescale=False, dilation=2)
            x_level_4_0 = cond_res_block(x_level_3_1, h, out_filters=2*self.nr_filters, rescale=True, dilation=4)
            x_level_4_1 = cond_res_block(x_level_4_0, h, out_filters=2*self.nr_filters, rescale=False, dilation=4)

            refine_0 = cond_refine_block([x_level_4_1], h, nr_filters=2*self.nr_filters, out_shape=nn.int_shape(x_level_4_1)[1:3])
            refine_1 = cond_refine_block([x_level_3_1, refine_0], h, nr_filters=2*self.nr_filters, out_shape=nn.int_shape(x_level_3_1)[1:3])
            refine_2 = cond_refine_block([x_level_2_1, refine_1], h, nr_filters=1*self.nr_filters, out_shape=nn.int_shape(x_level_2_1)[1:3])
            refine_3 = cond_refine_block([x_level_1_1, refine_2], h, nr_filters=1*self.nr_filters, out_shape=nn.int_shape(x_level_1_1)[1:3], end=True)
            

            out = self.normalizer(refine_3, h)
            out = self.nonlinearity(out)
            out = nn.conv2d_plus(out, num_filters=self.chns, nonlinearity=None)
            
            self.counters = {} # reset counters

            return out


from scipy.fftpack import fftshift, ifftshift, ifft2, fft2
import numpy as np

def mfft2(x, dims, axis=(0,1), center=False):
    nx, ny = dims
    if center:
        return ifftshift(fft2(fftshift(x), axes=axis))/np.sqrt(nx*ny)
    else:
        return fftshift(fft2(fftshift(x), axes=axis))/np.sqrt(nx*ny)

def mifft2(x, dims, axis=(0,1), center=False):
    nx, ny = dims
    if center:
        return fftshift(ifft2(ifftshift(x), axes=axis))*np.sqrt(nx*ny)
    else:
        return fftshift(ifft2(fftshift(x), axes=axis))*np.sqrt(nx*ny)


def A_cart(img, coilsen, mask, shape, axis=(0,1), center=False):
    """
    forward cartesian A 
    """
    coil_img = coilsen*img[..., np.newaxis]
    kspace = mfft2(coil_img, shape, axis=axis, center=center)
    kspace = np.multiply(kspace, mask[...,np.newaxis])
    return kspace

def AT_cart(kspace, coilsen, mask, shape, axis=(0,1), center=False):
    """
    adjoint cartesian AT
    coil dimension should always be the last
    """
    coil_img = mifft2(kspace*mask[...,np.newaxis], shape, axis=axis, center=center)
    coil_sum = np.sum(coil_img*np.conj(coilsen), axis=-1)
    return coil_sum

def sense_kernel_cart(AHy, img_k, coilsen, mask, shape, step_size=1, axis=(1,2), center=False):
    tmp = A_cart(img_k, coilsen, mask, shape, axis=axis, center=center)
    tmp = AT_cart(tmp, coilsen, mask, shape, axis=axis, center=center)
    img_k = img_k + step_size*(AHy - tmp)
    return img_k

def AHA(x, coilsen, mask, shape, axis=(1,2), center=False):
    tmp = A_cart(x, coilsen, mask, shape, axis, center)
    ret = AT_cart(tmp, coilsen, mask, shape, axis, center)
    return ret


from spreco.model import nn, blocks
import tensorflow.compat.v1 as tf
import numpy as np
from tf_slim import arg_scope

class unet():

    def __init__(self, name, config, chns):

        super().__init__(name,
                        config,
                        tf.make_template('forward', self.body),
                        chns, nn.get_nonlinearity(config['nonlinearity']))
        self.counters    = {}

        self.nr_filters = self.config['nr_filters']
        self.attention  = self.config['attention']
        self.ch_factor  = self.config['ch_factor']
        self.nr_resnet  = self.config['nr_resnet']

    def body(self, x, t):

        with arg_scope([nn.conv2d_plus, blocks.cond_resnet, nn.self_attention, nn.downsample, nn.upsample, nn.nin], nonlinearity=self.nonlinearity, counters=self.counters, normalizer=None):

            # downstream
            temb = nn.get_timestep_embedding(t, embedding_size=self.nr_filters)
            temb = nn.nin(temb, self.nr_filters * 4, self.nonlinearity)
            temb = nn.nin(temb, self.nr_filters * 4)

            blks = []

            x = nn.conv2d_plus(x, self.nr_filters, nonlinearity=None)
            blks.append(x)

            for level, (factor, attn) in enumerate(zip(self.ch_factor, self.attention)):

                current_filters = factor*self.nr_filters
                for i in range(self.nr_resnet):
                    x = blocks.cond_resnet(x, temb, current_filters, rescale=False, rescale_with_conv=True)
                    if attn:
                        x = nn.self_attention(x, qk_chns=current_filters, v_chns=current_filters)
                    blks.append(x)
                if level != len(self.ch_factor)-1:
                    x = nn.downsample(x, with_conv=True)
                    blks.append(x)

            # bottom
            x = blocks.cond_resnet(x, temb, current_filters, rescale=False)
            x = nn.self_attention(x, current_filters, current_filters)
            x = blocks.cond_resnet(x, temb, current_filters, rescale=False)
            
            # upstream
            for level, (factor, attn) in enumerate(zip(reversed(self.ch_factor), reversed(self.attention))):

                current_filters = factor*self.nr_filters
                for i in range(self.nr_resnet+1):

                    x = blocks.cond_resnet(tf.concat([x, blks.pop()], axis=-1), temb, current_filters, rescale=False, rescale_with_conv=True)
                    if attn:
                        x = nn.self_attention(x, qk_chns=current_filters, v_chns=current_filters)

                if level!= len(self.ch_factor)-1:
                    x = nn.upsample(x, x.shape[1:3], 2, current_filters, True)
            x = self.nonlinearity(x)
            x = nn.conv2d_plus(x, num_filters=self.out_chns, nonlinearity=None)
            self.counters = {}
            return x


from spreco.dataflow.base import RNGDataFlow
from spreco.dataflow.common import BatchData
from spreco.dataflow.parallel_map import MultiThreadMapData
from spreco.common import utils

import numpy as np

class cfl_pipe(RNGDataFlow):

    def __init__(self, files, shuffle):
        self._size   = len(files)
        self.files   = files
        self.shuffle = shuffle
    
    def __len__(self):
        return len(self.files)
    
    def __iter__(self):
        idxs = np.arange(self._size)
        if self.shuffle:
            self.rng.shuffle(idxs)
        
        for idx in idxs:
            fname = self.files[idx]
            yield fname

train_files = utils.read_filelist('test.txt')
dp = cfl_pipe(train_files, True)
def map_f(x):
    return x[:-4]
dp = MultiThreadMapData(dp, 10, map_f,  buffer_size=300, strict=True)
dp = BatchData(dp, 4, use_list=True)

dp.reset_state()
for f in dp:
    print(f)


from . import utils
import numpy as np
from functools import partial
from copy import deepcopy

class DATA_CHNS():
    """

    """
    MAG  = 'MAG'  # for magnitude images
    CPLX = 'CPLX' # for images with real and imaginary channels 
    RGB  = 'RGB'  # for images with red, green, blue channnels

    @staticmethod
    def get():
        return [
            DATA_CHNS.MAG,
            DATA_CHNS.CPLX,
            DATA_CHNS.RGB
        ]

class MODELS():
    """

    """
    NCSN         = 'NCSN'
    SDE          = 'SDE'
    PIXELCNN     = 'PIXELCNN'
    
    @staticmethod
    def get():
        return [
            MODELS.NCSN,
            MODELS.SDE,
            MODELS.PIXELCNN
        ]


class parts():
    """
    class to get parts_func for building pipe
    """

    dict_func = {}
    
    dict_func['abs']         = abs
    dict_func['squeeze']     = np.squeeze
    dict_func['slice_image'] = utils.slice_image  # args -> (inp, shape)
    dict_func['npz_loader']  = utils.npz_loader   # args -> (filename, key='rss')
    dict_func['cplx2float']  = utils.cplx2float   # args -> (cplx_in)
    dict_func['float2cplx']  = utils.float2cplx   # args -> (float_in)
    dict_func['cfl_loader']  = utils.readcfl      # args -> (filename)
    dict_func['normalize_with_max'] = utils.normalize_with_max    # args -> (x, axis=(0,1), data_chns='CPLX')
    dict_func['transform']          = utils.transform             # args -> (inp)
    dict_func['random_rescale']     = utils.random_rescale        # args -> (x, vmin, vmax)
    dict_func['randint']            = utils.randint               # args -> (x, nr_levels, dtype)
    dict_func['randfloat']          = utils.randfloat             # args -> (x, lower=0, upper=1, eps, T) interval [eps, T)
    dict_func['affine']             = utils.affine                # args -> (x, a, b) 
    dict_func['expand_dims']        = utils.expand_dims           # args -> (x, axis) 

    @staticmethod
    def parse(parts_ll):
        cl = deepcopy(parts_ll)
        ret = []
        for parts_l in cl:
            l = []
            for part in parts_l:
                func_key = part.pop('func', None)

                if func_key is None:
                    raise Exception("check the definition of part in config file")
                else:
                    try:
                        func = parts.dict_func[func_key]
                    except:
                        raise Exception("the function %s is not defined"% func_key)

                l.append(partial(func, **part))
            ret.append(l)
        return ret


import numpy as np

def gen_mask_1D(ratio=0.1,center=20, ph=256, fe=256):
    """
    generate undersampling mask along 1 dimension
    Args:

    ratio: sampling ratio
    center: center lines retained
    fe: frequency encoding
    ph: phase encoding lines

    Returns:
    mask
    """
    k = int(round(ph*ratio)/2.0)
    ma = np.zeros(ph)
    ri = np.random.choice(int(ph/2-center/2), k, replace=False)
    ma[ri] = 1
    ri = np.random.choice(int(ph/2-center/2), k, replace=False)
    ma[ri+int(ph/2+center/2)] = 1
    ma[int(ph/2-center/2): int(ph/2+center/2)] = 1
    mask = np.tile(ma, [fe, 1])
    return mask

def gen_mask(ny, factor):
    center_fraction = (32 // 4) / 100
    num_low_freqs = int(ny * center_fraction)
    acs_lim = (ny-num_low_freqs+1)//2
    num_high_freqs = ny // factor - num_low_freqs
    high_freqs_spacing = (ny - num_low_freqs) // num_high_freqs
    mask_offset = int(np.random.uniform(low=0, high=high_freqs_spacing))
    high_freqs_location = np.arange(mask_offset, ny, high_freqs_spacing)
    low_freqs_location=np.arange(acs_lim, acs_lim+num_low_freqs)
    mask_locations = np.concatenate([high_freqs_location, low_freqs_location])
    mask = np.zeros([ny], dtype=np.int32)
    mask[mask_locations]=1
    return mask


def gen_mask_2D( nx, ny, center_r = 20, undersampling = 0.3 ):
    #create undersampling mask
    k = int(round(nx*ny*undersampling)) #undersampling
    ri = np.random.choice(nx*ny,k,replace=False) #index for undersampling
    ma = np.zeros(nx*ny) #initialize an all zero vector
    ma[ri] = 1 #set sampled data points to 1
    mask = ma.reshape((nx,ny))

    # center k-space index range
    if center_r > 0:

        cx = np.int(nx/2)
        cy = np.int(ny/2)

        cxr_b = round(cx-center_r)
        cxr_e = round(cx+center_r+1)
        cyr_b = round(cy-center_r)
        cyr_e = round(cy+center_r+1)

        mask[cxr_b:cxr_e, cyr_b:cyr_e] = 1. #center k-space is fully sampled

    return mask

def gen_mask_2d_equally_skip(factor_x=2, factor_y=2, center=20, fe=256, ph=256):
    """

    """
    p = np.zeros((fe, ph))
    
    grid = np.meshgrid(np.arange(0, fe, factor_x), np.arange(0, ph, factor_y))
    p[tuple(grid)] = 1

    if center > 0:
        cx = np.int(fe/2)
        cy = np.int(ph/2)

        cxr_b = round(cx-center//2)
        cxr_e = round(cx+center//2+1)
        cyr_b = round(cy-center//2)
        cyr_e = round(cy+center//2+1)

        p[cxr_b:cxr_e, cyr_b:cyr_e] = 1. #center k-space is fully sampled

    return p


from spreco.model import nn
from spreco.model import utils

import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tf_slim import add_arg_scope

@add_arg_scope
def cond_resnet(x, h, out_filters, nonlinearity, normalizer, rescale=False, rescale_with_conv=True, **kwargs):
    """
    resnet block
    out_filters is output_dims/feature
    """

    in_filters = nn.int_shape(x)[-1]
    x_skip = x

    if normalizer is not None:
        x = normalizer(x) 
    x = nonlinearity(x)

    if rescale:
        x = nn.conv2d_plus(x, in_filters, nonlinearity=None, scope='cond_res', **kwargs)
    else:
        x = nn.conv2d_plus(x, out_filters, nonlinearity=None, scope='cond_res', **kwargs)

    if normalizer is not None:
        x = normalizer(x)
    x = nonlinearity(x)

    x = nn.conv2d_plus(x, out_filters, nonlinearity=None, scope='cond_res', **kwargs)
    if rescale:
        x = tf.nn.avg_pool2d(x, ksize=(1,2,2,1), strides=(1,2,2,1), padding='SAME')
    
    if out_filters == in_filters and not rescale: # the shape of current variable remains the same as 
        shortcut = x_skip
    else:
        if rescale:
            if rescale_with_conv:
                shortcut = nn.conv2d_plus(x_skip, out_filters, filter_size=[3, 3], stride=[2,2])
            else:
                shortcut = tf.nn.avg_pool2d(x_skip, ksize=(1,2,2,1), strides=(1,2,2,1), padding='SAME')
        else:
            shortcut = nn.conv2d_plus(x_skip, out_filters)   # increase filters inside block

    if h is not None:
        shortcut = shortcut + tf.expand_dims(tf.expand_dims(nn.nin(h, out_filters), 1), 1)

    return shortcut + x

@add_arg_scope
def encoder(x, h, z_filters,
            num_filters=64,
            num_resnet=2,
            ch_factor=[1,2,4,8],
            attention=[False, False, True, True],
            rescale_with_conv=True,
            normalizer=nn.instance_norm, nonlinearity=None, **kwargs):
    """
    encode input x into latent space z

    Args:
      x: input
      h: conditional info
      z_filters: the number of features in latent space
      num_filter: the number of features get to the 1st block
      num_resnet: the number of resnet contained within a single block
      ch_factor: the list contains the factor by which the num of features increase
      attention: the list indicate which block has self-attention module
      rescale_with_conv: the bool value tell rescale with convolution or pooling
      normalizer: the normalizer
    """

    blks=[]

    x = nn.conv2d_plus(x, num_filters, nonlinearity=None)
    blks.append(x)

    for level, (factor, attn) in enumerate(zip(ch_factor, attention)):

        current_filters = factor*num_filters
        
        for i in range(num_resnet):

            rescale = False
            if i == num_resnet-1 and level != len(ch_factor)-1:
                rescale=True
            x = cond_resnet(x, h, current_filters, rescale=rescale, rescale_with_conv=rescale_with_conv)
            if attn:
                x = nn.self_attention(x, qk_chns=current_filters, v_chns=current_filters)
            blks.append(x)
    
    x = normalizer(x)
    x = nonlinearity(x)
    z = nn.conv2d_plus(x, z_filters, nonlinearity=None, scope='latent_space')

    return z


@add_arg_scope
def decoder(x, h, out_filters, 
            num_filters=64,
            num_resnet=2,
            ch_factor=[1,2,4,8],
            attention=[True, True, False, False],
            normalizer=nn.instance_norm, nonlinearity=None, **kwargs):
    """
    decode latent space z into x

    Args:
      z, the latent space
      h, the conditional information
      out_filters, the number of channel of output
      num_resnet, the number of resnet that contained within a single block
      ch_factor, the list contains factor by which the number of features increases
      attention, the list indicate which block has self-attention module
      normalier, the default is instance normalizer
    """

    for level, (factor, attn) in enumerate(zip(ch_factor, attention)):

        current_filters = num_filters*factor

        for i in range(num_resnet):

            x = cond_resnet(x, h, current_filters)
            if attn:
                x = nn.self_attention(x, qk_chns=current_filters, v_chns=current_filters)
            if i == num_resnet-1 and level != len(ch_factor)-1:
                x = nn.upsample_2(x, x.shape[1:3], 2, current_filters)
                #x = nn.upsample(x, x.shape[1:3], 2, current_filters, with_conv=False)

    x = normalizer(x)
    x = nonlinearity(x)
    x = nn.conv2d_plus(x, out_filters, nonlinearity=None, scope='output')

    return x

@add_arg_scope
def transformer_block(x, past, num_heads, num_embedings, **kwargs):
    y = nn.nlp_norm(x, scope='transformer_layernorm')
    a, present = nn.causal_attention(y, past, num_heads, num_embedings)
    x = x + a 
    y = nn.nlp_norm(x, scope='transformer_layernorm')
    m = nn.mlp(y, num_embedings*4)
    return x+m, present

@add_arg_scope
def discriminator(x, num_filters, num_layers, norm_layer, **kwargs):

    x = nn.conv2d_plus(x, num_filters, filter_size=[4,4], stride=[2,2])
    x = tf.nn.leaky_relu(x)

    nf_mult = 1

    for n in range(1, num_layers):
        nf_mult = min(2**n, 8)
        x = nn.conv2d_plus(x, num_filters*nf_mult, filter_size=[4,4], stride=[2,2])
        x = norm_layer(x)
        x = tf.nn.leaky_relu(x)

    nf_mult = min(2**n, 8)
    x = nn.conv2d_plus(x, num_filters*nf_mult, filter_size=[4,4])
    x = norm_layer(x)
    x = tf.nn.leaky_relu(x)

    nf_mult = min(2**n, 8)
    x = nn.conv2d_plus(x, 1, filter_size=[4,4])

    return x


import os
from datetime import datetime
from termcolor import colored
from contextlib import redirect_stdout

import tensorflow.compat.v1 as tf
import tempfile 
import yaml

import subprocess as sp
import numpy as np
from skimage.metrics import structural_similarity, peak_signal_noise_ratio
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt


def float2cplx(float_in):
    return np.array(float_in[...,0]+1.0j*float_in[...,1], dtype='complex64')

def cplx2float(cplx_in):
    return np.array(np.stack((cplx_in.real, cplx_in.imag), axis=-1), dtype='float32')

def log_to(file, vars, mode="a", clear=False, end='\n', prefix=None):
    if clear:
        open(file, "w").close()
    else:
        with open(file, mode) as f:
            with redirect_stdout(f):
                
                for var in vars:
                    if prefix is not None:
                        print(prefix, end='')
                    print(var, end=end)

def save_img(img, path, vmin=0., vmax=1., cmap='gray', interpolation=None):
    """
    print images to pdf and png without white margin

    Args:
    img: image arrays
    path: saving path
    """
    plt.imshow(img, cmap=cmap, interpolation=interpolation, vmin=vmin, vmax=vmax)
    plt.axis('off')
    plt.savefig(path, bbox_inches='tight', pad_inches = 0)
    plt.savefig(path+'.pdf', bbox_inches='tight', pad_inches = 0)
    plt.close()

def load_config(path):
    """
    load configuration defined with yaml file
    """
    with open(path, "r") as config_file:
        config = yaml.load(config_file, Loader=yaml.FullLoader)
    return config

def save_config(x,path):
    with open(os.path.join(path, 'config.yaml'), 'w') as yaml_file:
        yaml.dump(x, yaml_file, default_flow_style=False, sort_keys=False)
    
def create_folder(save_path, time=True):
    """
    create folder for logs
    """
    if time:
        log_path = os.path.join(save_path, datetime.now().strftime("%Y%m%d-%H%M%S"))
    else:
        log_path = save_path

    if not os.path.exists(log_path):
        os.makedirs(log_path)
    return log_path

def check_out(cmd, split=True):
    """ utility to check_out terminal command and return the output"""

    strs = sp.check_output(cmd, shell=True).decode()

    if split:
        split_strs = strs.split('\n')[:-1]
    return split_strs

def find_files(path, pattern):
    cmd = "find " + path + " -type f -name " + pattern
    return check_out(cmd)

def read_filelist(filename):
    with open(filename) as f:
        lines = [line.rstrip() for line in f]
        return lines

def list_data_sys(path, suffix, ex_pattern=None):
    """
    list all the files under path with given suffix

    Args:

    path: the folder to be listed
    suffix: suffix of file to be listed
    ex_pattern: pattern for excluding files

    Returns:
    
    array of the files' path
    """

    cmd = 'ls '
    cmd = cmd + path + '/*.' + suffix

    if ex_pattern is not None:
        cmd = cmd + ' | ' + 'grep -v ' + ex_pattern
    
    strs = sp.check_output(cmd, shell=True).decode()
    files = strs.split('\n')[:-1]

    return files

def slice_image(inp, shape):
    """
    slice image into pieces

    Args:
    inp: (nx, ny, _)
    shape: 

    Returns:
    
    """
    nx = ny = None
    if len(inp.shape) == 3:
        nx, ny, _ = inp.shape
    elif len(inp.shape) == 2:
        nx, ny = inp.shape
    else:
        print(inp.shape)
        raise Exception("please check the shape of input")

    if len(shape) == 3:
        sx, sy, _ = shape
    elif len(shape) == 2:
        sx, sy = shape
    else:
        raise Exception("Please check the given shape")

    steps_x = int(np.ceil(float(nx)/sx))
    steps_y = int(np.ceil(float(ny)/sy))

    total = steps_x*steps_y
    pieces = np.zeros([total] + shape, dtype=inp.dtype)
     
    for x in range(steps_x):
        
        if x == (steps_x-1):
            bx = nx-sx
            ex = nx
        else:
            bx = x*sx
            ex = bx + sx

        for y in range(steps_y):

            if y == (steps_y-1):
                by = ny-sy
                ey = ny
            else:
                by = y*sy
                ey = by + sy

            pieces[x*steps_y+y, ...] = np.reshape(inp[bx:ex, by:ey], shape)
    return pieces

def npz_loader(filename, key='rss'):
    tmp = np.load(filename)
    if key not in tmp.keys():
        info = "File loading failed, key %s doesn't match or exist!"%key
        #color_print(info)
        raise Exception(info)
    else:
        image = tmp[key]
    return image

def normalize_with_max(x, axis=(0,1), data_chns='CPLX'):
    """
    x is complex value
    x = x/(max(abs(x)))
    """
    scalor = np.max(abs(x), axis)

    if data_chns == 'CPLX':
        normalized_x = cplx2float(x/scalor)
        return normalized_x
    
    if data_chns == 'MAG':
        normalized_x = abs(x/scalor)
        return normalized_x

def color_print(strs, color='red', bold=True):
    if bold:
        print(colored(strs, color, attrs=['bold']))
    else:
        print(colored(strs, color))

def readcfl(name):
    """
    read cfl file
    """

    # get dims from .hdr
    h = open(name + ".hdr", "r")
    h.readline() # skip
    l = h.readline()
    h.close()
    dims = [int(i) for i in l.split( )]

    # remove singleton dimensions from the end
    n = np.prod(dims)
    dims_prod = np.cumprod(dims)
    dims = dims[:np.searchsorted(dims_prod, n)+1]

    # load data and reshape into dims
    d = open(name + ".cfl", "r")
    a = np.fromfile(d, dtype=np.complex64, count=n)
    d.close()
    return a.reshape(dims, order='F') # column-major

def writecfl(name, array):
    """
    write cfl file
    """

    h = open(name + ".hdr", "w")
    h.write('# Dimensions\n')
    for i in (array.shape):
            h.write("%d " % i)
    h.write('\n')
    h.close()
    d = open(name + ".cfl", "w")
    array.T.astype(np.complex64).tofile(d) # tranpose for column-major order
    d.close()

def transform(inp):
    """
    function for data augmentation
    TODO: zoom in, zoom out
    """
    k = np.random.choice(12)
    if k == 0:
        return inp
    if k == 1:    
        return np.flipud(inp)
    if k == 2:
        return np.fliplr(inp)
    if k == 3:
        return np.rot90(inp, 1)
    if k == 4:
        return np.rot90(inp, 2)
    if k == 5:
        return np.rot90(inp, 3)
    if k == 6:
        return np.rot90(np.flipud(inp), 1)
    if k == 7:
        return np.rot90(np.flipud(inp), 2)
    if k == 8:
        return np.rot90(np.flipud(inp), 3)
    if k == 9:
        return np.rot90(np.fliplr(inp), 1)
    if k == 10:
        return np.rot90(np.fliplr(inp), 2)
    if k == 11:
        return np.rot90(np.fliplr(inp), 3)


def get_timestamp():
    return datetime.now().strftime("%Y%m%d-%H%M%S")

def export_model(saver, sess, path, name, as_text=False, gpu_id=None):
    saver.save(sess, os.path.join(path, name))
    tf.train.write_graph(sess.graph, path, name+'.pb', as_text)
    if gpu_id is not None:
        with open(os.path.join(path, name+'_gpu_id'), 'w+') as fs:
            for i in range(len(gpu_id)):
                fs.write(gpu_id[i])
                fs.write('\t')

def random_rescale(x, vmin=0.7, vmax=1.):
    return x*round(np.random.uniform(vmin,vmax),4)

def randint(x, nr_levels, dtype='int32'):
    # x is a dummy arg
    return np.random.randint(0, nr_levels, (1), dtype=dtype)

def randfloat(x, eps, T):
    # x is a dummy arg
    return np.random.uniform(eps, T, size=(1))

def to_dict(val, key='inputs'):
    return {key: val}

def gaussian_noise(shape, mu=0, sigma=1):
        return np.random.normal(mu, sigma, shape).astype(np.float32)
    
def gaussian_noise_bivariate(shape, mu=[0, 0], sigma=[[1, 0],[0, 1]]):
        tmp = np.random.multivariate_normal(mu, sigma, np.prod(shape)).astype(np.float32)
        return float2cplx(np.reshape(tmp, [shape[0], shape[1], 2]))

def affine(x, a, b):
    return (x-a)/b

def expand_dims(x, axis):
    return np.expand_dims(x, axis)

def bart(nargout, cmd, *args, return_str=False):
    """
    call bart from system command line
    """
    if type(nargout) != int or nargout < 0:
        print("Usage: bart(<nargout>, <command>, <arguements...>)")
        return None

    name = tempfile.NamedTemporaryFile().name

    nargin = len(args)
    infiles = [name + 'in' + str(idx) for idx in range(nargin)]
    in_str = ' '.join(infiles)

    for idx in range(nargin):
        writecfl(infiles[idx], args[idx])

    outfiles = [name + 'out' + str(idx) for idx in range(nargout)]
    out_str = ' '.join(outfiles)

    shell_str = 'bart ' + cmd + ' ' + in_str + ' ' + out_str
    print(shell_str)
    if not return_str:
        ERR = os.system(shell_str)
    else:
        try:
            strs = sp.check_output(shell_str, shell=True).decode()
            return strs
        except:
            ERR = True


    for elm in infiles:
        if os.path.isfile(elm + '.cfl'):
            os.remove(elm + '.cfl')
        if os.path.isfile(elm + '.hdr'):
            os.remove(elm + '.hdr')

    output = []
    for idx in range(nargout):
        elm = outfiles[idx]
        if not ERR:
            output.append(readcfl(elm))
        if os.path.isfile(elm + '.cfl'):
            os.remove(elm + '.cfl')
        if os.path.isfile(elm + '.hdr'):
            os.remove(elm + '.hdr')

    if ERR:
        print("Make sureyou install bart properly")
        raise Exception("Command exited with an error.")

    if nargout == 1:
        output = output[0]

    return output

def print_parameters(file, mode="a"):
    """
    print the trainable parameters info of tf graph
    """
    total_parameters = 0
    with open(file, mode) as f:
        with redirect_stdout(f):
            
            for variable in tf.trainable_variables():
                # shape is an array of tf.Dimension
                print("========layer=======")
                print(variable)
                shape = variable.get_shape()
                print("shape->",shape)
                print("shape->len",len(shape))
                variable_parameters = 1
                for dim in shape:
                    print("dim->", dim)
                    variable_parameters *= dim
                print(variable_parameters)
                total_parameters += variable_parameters
            print(total_parameters)

def norm_to_uint(inp, bit=8):
    maximum = np.max(inp)
    out = inp/maximum*np.power(2., bit)
    
    tp = np.uint8
    if bit == 16:
        tp= np.uint16
    if bit == 32:
        tp= np.uint16
    return out.astype(tp)

def psnr(img1, img2, bit=16, tobit=False):
    """
    calculate peak SNR, img1-true, img2-test
    """
    pixel_max = np.max(img2)
    
    if tobit:
        img1 = norm_to_uint(img1, bit)
        img2 = norm_to_uint(img2, bit)
        pixel_max = np.power(2., bit) - 1

    img1 = img1.astype(np.float64)
    img2 = img2.astype(np.float64)
    pixel_max = pixel_max.astype(np.float64)

    return peak_signal_noise_ratio(img1, img2, data_range=pixel_max)


def ssim(img1, img2, bit=16, tobit=False):
    """
    calcualte similarity index between img1 and img2
    """
    
    scale = np.max(img2)
    if tobit:
        img1 = norm_to_uint(img1, bit)
        img2 = norm_to_uint(img2, bit)  
        scale = np.power(2., bit) - 1 
    img1.astype(np.float64)
    img2.astype(np.float64)
    scale.astype(np.float64)
    return structural_similarity(img1, img2, data_range=scale)

def get_lr(step, lr, warmup_steps=None, hidden_size=None):
    """
    learning rate scheduler
    """
    if warmup_steps is not None and hidden_size is not None:
        lr_base = lr * 0.002 # for Adam correction
        ret = 5000. * hidden_size ** (-0.5) * \
          np.min([(step + 1) * warmup_steps ** (-1.5), (step + 1) ** (-0.5)])
        return ret * lr_base
    else:
        return lr


import tensorflow as tf
import numpy as np

class Gaussian_Mixture():
    
    def __init__ (self, dim, mix_prob=[0.7, 0.2, 0.1]):

        self.mix_prob = mix_prob
        self.means    = tf.stack([5*tf.ones(dim), -5*tf.ones(dim), [5,-5]], axis=0)
        self.sigma    = 1.

    def sample(self, nr_samples, sigma=1.):
        mix_idx = tf.random.categorical(tf.math.log([self.mix_prob]), nr_samples)
        means   = tf.gather(self.means, mix_idx)
        return tf.random.normal(means.shape)*sigma + means

    def log_prob(self, samples, sigma=1.):
        logps = []
        for i in range(len(self.mix_prob)):
            logps.append((-tf.reduce_sum((samples - self.means[i]) ** 2, axis=-1) / (2 * sigma ** 2) - 0.5 * tf.math.log(
                2 * np.pi * sigma ** 2)) + tf.math.log(self.mix_prob[i]))
        logp = tf.reduce_logsumexp(tf.stack(logps, axis=0), axis=0)
        return logp

    def score(self, samples, sigma=1.):
        with tf.GradientTape() as g:
            g.watch(samples)
            log_probs = tf.reduce_sum(self.log_prob(samples, sigma))
            return g.gradient(log_probs, samples)




from spreco.model import nn
from spreco.model.refine_net import cond_refine_net
from spreco.common.custom_adam import AdamOptimizer

import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

class ncsn():
    """
    noise conditional score networks
    Generative modeling by estimating gradients of the data distribution arXiv: 1907.05600
    """
    def __init__(self, config):

        self.config       = config
        self.net          = cond_refine_net(config)
        self.begin_sigma  = config['begin_sigma']
        self.end_sigma    = config['end_sigma']
        self.nr_levels    = config['nr_levels']
        self.anneal_power = config['anneal_power']
        self.sigmas       = tf.exp(tf.linspace(tf.log(self.begin_sigma), tf.log(self.end_sigma), self.nr_levels))
        self.inputs       = [tf.placeholder(tf.float32, 
                               shape=[self.config['batch_size']]+self.config['input_shape'], name="input_%d"%i
                               ) for i in range(self.config['nr_gpu'])]
        self.h            = [tf.placeholder(tf.int32,
                               shape=[self.config['batch_size']]) for _ in range(config['nr_gpu'])]
        self.ins_outs     = {'inputs': self.inputs, 'h': self.h}
        self.learning_rate  = tf.placeholder(tf.float32, shape=[])
        self.seed         = config['seed']
        
    
    def denoise_score_matching(self, x):

        shape     = nn.int_shape(x)
        perturb_x = x + tf.random_normal(shape, mean=0.0, stddev=self.begin_sigma, seed=self.seed)
        target    = -1/(self.begin_sigma**2) * (perturb_x - x)
        score     = self.net.forward(perturb_x)
        tmp       = tf.reshape(score-target, (shape[0], -1))

        loss      = 1/2 * tf.reduce_mean(tf.reduce_sum(tmp**2, axis=-1))
        return loss
    
    def anneal_denoise_score_matching(self, x, h):
        
        shape     = nn.int_shape(x)
        sigs      = tf.gather(self.sigmas, h)[:, tf.newaxis, tf.newaxis, tf.newaxis]
        perturb_x = x + tf.random_normal(shape, seed=self.seed)*sigs
        target    = -1./(sigs**2) * (perturb_x - x)
        score     = self.net.forward(perturb_x, h)
        tmp       = tf.reshape(score-target, (shape[0], -1))

        loss      = tf.reduce_mean(1/2. * (tf.reduce_sum(tmp**2, axis=-1))*tf.squeeze(sigs)**self.anneal_power)
        return loss
    
    def anneal_denoise_score_matching_v2(self, x, h):
        """
        condition term h is not required
        """
        shape     = nn.int_shape(x)
        sigs      = tf.gather(self.sigmas, h)[:, tf.newaxis, tf.newaxis, tf.newaxis]
        perturb_x = x + tf.random_normal(shape, seed=self.seed)*sigs
        target    = -1./(sigs**2) * (perturb_x - x)
        score     = self.net.forward(perturb_x, [0]*shape[0])
        score     = score / sigs # Parameterize the NCSN with s(x, σ) = s(x)/σ 
        tmp       = tf.reshape(score-target, (shape[0], -1))

        loss      = tf.reduce_mean(1/2. * (tf.reduce_sum(tmp**2, axis=-1))*tf.squeeze(sigs)**self.anneal_power)
        return loss
    
    def prep(self, export=True):
        """make preparation for training objective"""

        if not export:
            _ = self.anneal_denoise_score_matching(self.inputs[0], self.h[0])
            all_params = tf.trainable_variables()

            loss      = []
            grads     = []
            loss_test = []

            optimizer = AdamOptimizer(self.learning_rate, beta1=0.9, beta2=0.999)
        
            for i in range(self.config['nr_gpu']):
                with tf.device('/gpu:%d'%i):
                    # train
                    loss.append(self.anneal_denoise_score_matching(self.inputs[i], self.h[i]))

                    gvs = optimizer.compute_gradients(loss[-1], all_params)
                    gvs = [(k, v) for (k, v) in gvs if k is not None]
                    grads.append(gvs)

                    # test
                    loss_test.append(self.anneal_denoise_score_matching(self.inputs[i], self.h[i]))
            
            with tf.device('/gpu:0'):
                for i in range(1, self.config['nr_gpu']):
                    loss[0] += loss[i]
                    loss_test[0] += loss_test[i]
            
            grads_avg = optimizer.average_gradients(grads)

            self.train_op = optimizer.apply_gradients(grads_avg)
            self.loss_train = loss[0]/self.config['nr_gpu']
            self.loss_test = loss_test[0]/self.config['nr_gpu']

        else:
            _ = self.anneal_denoise_score_matching(self.inputs[0], self.h[0])
            all_params = tf.trainable_variables()



import tensorflow as tf
from tensorflow.keras.layers import Dense
from tensorflow.keras import Model
import numpy as np


class net(Model):
    def __init__(self, config):
        super(net, self).__init__()
        self.d1 = Dense(config['num_units'], activation=config['nonlinearity'])
        self.d2 = Dense(config['num_units'], activation=config['nonlinearity'])
        self.d3 = Dense(2)

    def call(self, x):
        x = self.d1(x)
        x = self.d2(x)
        x = self.d3(x)
        return x


class toy():
    """
    the toy model to demonstrate how samples are simulated from posterior
    """
    def __init__(self, config):

        self.config       = config

        self.sigma_min     = config['sigma_min']
        self.sigma_max     = config['sigma_max']
        self.nr_levels     = config['nr_levels']
        self.sigmas        = tf.exp(tf.linspace(tf.math.log(self.sigma_max), self.sigma_min, self.nr_levels))
        self.seed          = config['seed']
        self.score_net     = net(config)


    @staticmethod
    def langevin_dynamics(score, x_k, lr=0.1, step=1000):
        samples_t=[]
        for i in range(step):
            samples_t.append(x_k)
            current_lr = lr
            x_k = x_k + current_lr / 2 * score(x_k)
            x_k = x_k + tf.random.normal(x_k.shape) * np.sqrt(current_lr)
        samples_t.append(x_k)
        return x_k, samples_t

    @staticmethod
    def anneal_langevin_dynamics(score, x_k, sigmas, lr=0.1, n_steps_each=100):
        samples_t=[]
        for sigma in sigmas:
            for i in range(n_steps_each):
                samples_t.append(x_k)
                current_lr = lr * (sigma / sigmas[-1]) ** 2
                x_k = x_k + current_lr / 2 * score(x_k, sigma)
                x_k = x_k + tf.random.normal(x_k.shape) * np.sqrt(current_lr)
        samples_t.append(x_k)

        return x_k, samples_t

    def score_estimation(self, samples):

        vectors = tf.random.normal(samples.shape)

        with tf.GradientTape() as g:
            g.watch(samples)
            grad1 = self.score_net(samples)
            gradv = tf.reduce_sum(grad1 * vectors)
            grad2 = g.gradient(gradv, samples)

        grad1 = tf.reshape(grad1, (samples.shape[0], -1))

        loss1 = tf.reduce_sum(grad1 * grad1, axis=-1) / 2.
        loss2 = tf.reduce_sum(tf.reshape(vectors * grad2, (samples.shape[0], -1)), axis=-1)
        loss = loss1 + loss2

        return tf.reduce_mean(loss)



from spreco.common.options import MODELS
from spreco.common import utils
from spreco.common.logger import logger
from spreco.common.utils import LambdaWarmUpCosineScheduler

import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

import os
import numpy as np
import time

class worker():
    """
    mode, 0: train
    mode, 1: inference
    mode, 2: export
    """

    def __init__(self, train_pipe, test_pipe, config, default_feed=True):

        self.train_pipe = train_pipe
        self.test_pipe  = test_pipe
        self.config     = config

        if train_pipe is not None:
            self.log_path   = utils.create_folder(config['log_folder'])
            utils.save_config(config, self.log_path)
            self.logger = logger(self.log_path)

        self.model      = None
        self.sess       = None

        self.global_step  = 0
        self.default_feed = default_feed


    def init_model(self, mode, feed_func=None, gpu_id=None):
        
        if self.config['model'] == MODELS.NCSN:
            tf.random.set_random_seed(self.config['seed'])
            from spreco.model.ncsn import ncsn as selected_class

        elif self.config['model'] == MODELS.SDE:
            from spreco.model.sde import sde as selected_class

        elif self.config['model'] == MODELS.PIXELCNN:
            from spreco.model.pixelcnn import pixelcnn as selected_class

        else:
            raise Exception("Currently, this model is not implemented!")

        if mode == 0:
            self.lr_scheduler = LambdaWarmUpCosineScheduler(self.config['lr_warm_up_steps'],
                                self.config['lr_min'], self.config['lr_max'], self.config['lr_start'], self.config['lr_max_decay_steps'])

        self.model = selected_class(self.config)

        self.model.init(mode)
        if not self.default_feed:
            self.feed_func=feed_func

        # set gpu
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        if gpu_id is None:
            os.environ["CUDA_VISIBLE_DEVICES"] = '-1'
        else: 
            os.environ["CUDA_VISIBLE_DEVICES"] = gpu_id

    def mk_feed_dict(self, pipe, reps=None):
        """"utility to make feed dict"""
        if reps is None:
            reps = self.config['nr_gpu']

        keys = self.model.ins_outs.keys()
        elm = pipe.get_next()
        feed_dict ={}

        for key in keys:
            tmp = np.split(elm[key], self.config['nr_gpu'])
            feed_dict.update({self.model.ins_outs[key][i]: tmp[i] for i in range(reps)})

        return feed_dict

    def run_op(self, sess, ops, loss, feed_dict, is_training):
        
        if is_training:
            if type(ops) == list:
                l = []
                for op, l_op in zip(ops, loss):
                    l_tmp, _ = sess.run([l_op, op], feed_dict=feed_dict)
                    l.append(l_tmp)

            else:
                l, _ = sess.run([loss, ops], feed_dict=feed_dict)

            return l

        else:
            if type(loss) == list:
                l = []
                for l_op in loss:
                    l_tmp = sess.run(l_op, feed_dict=feed_dict)
                    l.append(l_tmp)

            else:
                l = sess.run(loss, feed_dict=feed_dict)

            return l

    def mk_kvs(self, keys, values):
        kvs = {}
        for key, value in zip(keys, values):
            if type(value) == list:
                for i, a in enumerate(value):
                    key_ = "%s_%d"%(key,i)
                    kvs[key_] = a
            else:
                kvs[key] = value
        return kvs

    def mk_info(self, epoch, t, loss_avg, printing, is_training=True):

        if is_training:
            info = "Epochs %d, time %ds, train loss: %s" % (epoch, t, ''.join(str(loss_avg)))
        else:
            info = "Epochs %d, time %ds, test loss: %s" % (epoch, t, ''.join(str(loss_avg)))

        if printing:
            print(info)

        return info

    def train_loop(self):

        # ready to go
        init_op    = tf.global_variables_initializer()
        saver      = tf.train.Saver(max_to_keep=self.config['max_keep'])
        gpu_config = tf.ConfigProto(allow_soft_placement=True)
        sess       = tf.Session(config=gpu_config)
        sess.run(init_op)

        utils.print_parameters(self.log_path+'/layer_info')
        train_loss = []
        test_loss  = []
        begin      = time.time()

        while self.train_pipe.get_epoch() < self.config['max_epochs']:

            if self.default_feed:
                feed_dict = self.mk_feed_dict(self.train_pipe)
            else:
                feed_dict = self.feed_func(self, self.train_pipe)

            learning_rate = self.lr_scheduler(self.global_step)

            feed_dict.update({self.model.learning_rate: learning_rate})

            l = self.run_op(sess, self.model.train_op, self.model.loss_train, feed_dict, is_training=True)
            self.global_step = self.global_step + 1

            kvs = self.mk_kvs(['train_loss', 'learning_rate'], [l, learning_rate])
            self.logger.writekvs(kvs)

            train_loss.append(l)

            if self.train_pipe.check_epoch():

                # one epoch is finised
                self.train_pipe.update_epoch()
                if self.train_pipe.get_epoch() % self.config["save_interval"] == 0:
                    saver.save(sess, os.path.join(self.log_path, self.config['saved_name']+'_'+str(self.train_pipe.get_epoch())))

                epoch = self.train_pipe.get_epoch()
                train_loss_avg = np.mean(train_loss, axis=0).tolist()

                info = self.mk_info(epoch, time.time()-begin, train_loss_avg, self.config['print_loss'])
                utils.log_to(os.path.join(self.log_path, 'loss'), [info], prefix="->")
                kvs = self.mk_kvs(['train_loss_avg'], [train_loss_avg])
                self.logger.writekvs(kvs, epoch)
                begin      = time.time()
                train_loss = []

                # run test
                while not self.test_pipe.check_epoch():
                    if self.default_feed:
                        feed_dict = self.mk_feed_dict(self.test_pipe)
                    else:
                        feed_dict = self.feed_func(self, self.test_pipe)

                    l = self.run_op(sess, None, self.model.loss_test, feed_dict=feed_dict, is_training=False)
                    test_loss.append(l)

                self.test_pipe.update_epoch()
                test_loss_avg = np.mean(test_loss, axis=0).tolist()
                info = self.mk_info(epoch, time.time()-begin, test_loss_avg, self.config['print_loss'], False)
                utils.log_to(os.path.join(self.log_path, 'loss'), [info], prefix="->")
                kvs = self.mk_kvs(['test_loss_avg'], [test_loss_avg])
                self.logger.writekvs(kvs, epoch)

                test_loss  = []
                begin      = time.time()

    def train(self, feed_func=None):
        self.train_pipe.start()
        self.test_pipe.start()
        self.init_model(mode=0, feed_func=feed_func, gpu_id=self.config['gpu_id'])
        self.train_loop()
        self.train_pipe.stop()
        self.test_pipe.stop()

    def inference(self, model_path, func, batch_size=None, gpu_id=None, **kwargs):
        """
        Restore the model located at model_path 
        Create the tf sess and model
        Func must take sess and model as two args
        Kwargs will be passed to func and which could be None
        """

        if self.model is None and self.sess is None:
            self.init_model(mode=1, batch_size=batch_size, gpu_id=gpu_id)
            _, sess, _ = self.restore(model_path)
            self.sess = sess

        return func(self.sess, self.model, **kwargs)

    def export(self, model_path, export_path, name, gpu_id='0'):

        self.init_model(mode=2, gpu_id=gpu_id)
        saver, sess, gpu_id = self.restore(model_path)
        utils.export_model(saver, sess, export_path, name, gpu_id=gpu_id)

    def restore(self, model_path):
        """restore the given model"""

        saver        = tf.train.Saver()
        gpu_options  = tf.GPUOptions(allow_growth=True, visible_device_list='0')
        config_proto = tf.ConfigProto(gpu_options=gpu_options)
        serialized   = config_proto.SerializeToString()
        gpu_id       = list(map(hex, serialized))
        sess         = tf.Session(config=config_proto)
        saver.restore(sess, model_path)
        return saver, sess, gpu_id


from spreco.model.utils import int_shape

import numpy as np 
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

def log_sum_exp(x):
    """ numerically stable log_sum_exp implementation that prevents overflow """
    axis = len(x.get_shape())-1
    m = tf.reduce_max(x, axis)
    m2 = tf.reduce_max(x, axis, keepdims=True)
    return m + tf.log(tf.reduce_sum(tf.exp(x-m2), axis))

def log_prob_from_logits(x):
    """ numerically stable log_softmax implementation that prevents overflow """
    axis = len(x.get_shape())-1
    m = tf.reduce_max(x, axis, keepdims=True)
    return x - m - tf.log(tf.reduce_sum(tf.exp(x-m), axis, keepdims=True))

def discretized_mix_logistic_loss(x,l,itg_interval=255.0, sum_all=True):
    """ log-likelihood for mixture of discretized logistics, assumes the data has been rescaled to [-1,1] interval """
    xs = int_shape(x) # true image (i.e. labels) to regress to, e.g. (B,32,32,3)
    ls = int_shape(l) # predicted distribution, e.g. (B,32,32,100)
    nr_mix = int(ls[-1] / 10) # here and below: unpacking the params of the mixture of logistics
    logit_probs = l[:,:,:,:nr_mix]
    l = tf.reshape(l[:,:,:,nr_mix:], xs + [nr_mix*3])
    means = l[:,:,:,:,:nr_mix]
    log_scales = tf.maximum(l[:,:,:,:,nr_mix:2*nr_mix], -7.)
    coeffs = tf.nn.tanh(l[:,:,:,:,2*nr_mix:3*nr_mix])
    x = tf.reshape(x, xs + [1]) + tf.zeros(xs + [nr_mix]) # here and below: getting the means and adjusting them based on preceding sub-pixels
    m2 = tf.reshape(means[:,:,:,1,:] + coeffs[:, :, :, 0, :] * x[:, :, :, 0, :], [xs[0],xs[1],xs[2],1,nr_mix])
    m3 = tf.reshape(means[:, :, :, 2, :] + coeffs[:, :, :, 1, :] * x[:, :, :, 0, :] + coeffs[:, :, :, 2, :] * x[:, :, :, 1, :], [xs[0],xs[1],xs[2],1,nr_mix])
    means = tf.concat([tf.reshape(means[:,:,:,0,:], [xs[0],xs[1],xs[2],1,nr_mix]), m2, m3],3)
    centered_x = x - means
    inv_stdv = tf.exp(-log_scales)
    plus_in = inv_stdv * (centered_x + 1./itg_interval)
    cdf_plus = tf.nn.sigmoid(plus_in)
    min_in = inv_stdv * (centered_x - 1./itg_interval)
    cdf_min = tf.nn.sigmoid(min_in)
    log_cdf_plus = plus_in - tf.nn.softplus(plus_in) # log probability for edge case of 0 (before scaling)
    log_one_minus_cdf_min = -tf.nn.softplus(min_in) # log probability for edge case of 255 (before scaling)
    cdf_delta = cdf_plus - cdf_min # probability for all other cases
    mid_in = inv_stdv * centered_x
    log_pdf_mid = mid_in - log_scales - 2.*tf.nn.softplus(mid_in) # log probability in the center of the bin, to be used in extreme cases (not actually used in our code)

    # now select the right output: left edge case, right edge case, normal case, extremely low prob case (doesn't actually happen for us)

    # this is what we are really doing, but using the robust version below for extreme cases in other applications and to avoid NaN issue with tf.select()
    # log_probs = tf.select(x < -0.999, log_cdf_plus, tf.select(x > 0.999, log_one_minus_cdf_min, tf.log(cdf_delta)))

    # robust version, that still works if probabilities are below 1e-5 (which never happens in our code)
    # tensorflow backpropagates through tf.select() by multiplying with zero instead of selecting: this requires use to use some ugly tricks to avoid potential NaNs
    # the 1e-12 in tf.maximum(cdf_delta, 1e-12) is never actually used as output, it's purely there to get around the tf.select() gradient issue
    # if the probability on a sub-pixel is below 1e-5, we use an approximation based on the assumption that the log-density is constant in the bin of the observed sub-pixel value
    log_probs = tf.where(x < -0.999, log_cdf_plus, tf.where(x > 0.999, log_one_minus_cdf_min, tf.where(cdf_delta > 1e-5, tf.log(tf.maximum(cdf_delta, 1e-12)), log_pdf_mid - np.log(127.5))))

    log_probs = tf.reduce_sum(log_probs,3) + log_prob_from_logits(logit_probs)
    if sum_all:
        return -tf.reduce_sum(log_sum_exp(log_probs))
    else:
        return -tf.reduce_sum(log_sum_exp(log_probs),[1,2])

def sample_from_discretized_mix_logistic(l,nr_mix):
    ls = int_shape(l)
    xs = ls[:-1] + [3]
    # unpack parameters
    logit_probs = l[:, :, :, :nr_mix]
    l = tf.reshape(l[:, :, :, nr_mix:], xs + [nr_mix*3])
    # sample mixture indicator from softmax
    sel = tf.one_hot(tf.argmax(logit_probs - tf.log(-tf.log(tf.random_uniform(logit_probs.get_shape(), minval=1e-5, maxval=1. - 1e-5))), 3), depth=nr_mix, dtype=tf.float32)
    sel = tf.reshape(sel, xs[:-1] + [1,nr_mix])
    # select logistic parameters
    means = tf.reduce_sum(l[:,:,:,:,:nr_mix]*sel,4)
    log_scales = tf.maximum(tf.reduce_sum(l[:,:,:,:,nr_mix:2*nr_mix]*sel,4), -7.)
    coeffs = tf.reduce_sum(tf.nn.tanh(l[:,:,:,:,2*nr_mix:3*nr_mix])*sel,4)
    # sample from logistic & clip to interval
    # we don't actually round to the nearest 8bit value when sampling
    u = tf.random_uniform(means.get_shape(), minval=1e-5, maxval=1. - 1e-5)
    x = means + tf.exp(log_scales)*(tf.log(u) - tf.log(1. - u))
    x0 = tf.minimum(tf.maximum(x[:,:,:,0], -1.), 1.)
    x1 = tf.minimum(tf.maximum(x[:,:,:,1] + coeffs[:,:,:,0]*x0, -1.), 1.)
    x2 = tf.minimum(tf.maximum(x[:,:,:,2] + coeffs[:,:,:,1]*x0 + coeffs[:,:,:,2]*x1, -1.), 1.)
    return tf.concat([tf.reshape(x0,xs[:-1]+[1]), tf.reshape(x1,xs[:-1]+[1]), tf.reshape(x2,xs[:-1]+[1])],3)


def discretized_mix_logistic_loss_2(x, l, itg_interval=255.0, sum_all=True, return_prob=False):
    """
    1. log-likelihood for mixture of discretized logistics, assumes the data has been rescaled to [-1,1] interval 
    2. this loss assume that real part is linearly dependent on the imaginary part
    """
    xs = int_shape(x) 
    ls = int_shape(l) 

    # unpacking the params of the mixture of logistics
    nr_mix = int(ls[-1] / 6) 
    logit_probs = l[:,:,:,:nr_mix]
    l = tf.reshape(l[:,:,:,nr_mix:], xs[:-1] + [5,nr_mix])
    means = l[:,:,:,:2,:]
    log_scales = tf.maximum(l[:,:,:,2:4,:], -7.)
    coeffs = tf.nn.tanh(l[:,:,:,4:,:])

    # getting the means and adjusting them based on preceding sub-pixels
    x = tf.reshape(x, xs + [1]) + tf.zeros(xs + [nr_mix]) 
    m2 = tf.reshape(means[:,:,:,1,:] + coeffs[:, :, :, 0, :] * x[:, :, :, 0, :], [xs[0],xs[1],xs[2],1,nr_mix])
    means = tf.concat([tf.reshape(means[:,:,:,0,:], [xs[0],xs[1],xs[2],1,nr_mix]), m2],3)
    centered_x = x - means
    inv_stdv = tf.exp(-log_scales)
    plus_in = inv_stdv * (centered_x + 1./itg_interval)
    cdf_plus = tf.nn.sigmoid(plus_in)
    min_in = inv_stdv * (centered_x - 1./itg_interval)
    cdf_min = tf.nn.sigmoid(min_in)
    log_cdf_plus = plus_in - tf.nn.softplus(plus_in) # log probability for edge case of 0 (before scaling)
    log_one_minus_cdf_min = -tf.nn.softplus(min_in) # log probability for edge case of 255 (before scaling)
    cdf_delta = cdf_plus - cdf_min # probability for all other cases
    mid_in = inv_stdv * centered_x
    log_pdf_mid = mid_in - log_scales - 2.*tf.nn.softplus(mid_in) # log probability in the center of the bin, to be used in extreme cases (not actually used in our code)

    # now select the right output: left edge case, right edge case, normal case, extremely low prob case (doesn't actually happen for us)

    # this is what we are really doing, but using the robust version below for extreme cases in other applications and to avoid NaN issue with tf.select()
    # log_probs = tf.select(x < -0.999, log_cdf_plus, tf.select(x > 0.999, log_one_minus_cdf_min, tf.log(cdf_delta)))

    # robust version, that still works if probabilities are below 1e-5 (which never happens in our code)
    # tensorflow backpropagates through tf.select() by multiplying with zero instead of selecting: this requires use to use some ugly tricks to avoid potential NaNs
    # the 1e-12 in tf.maximum(cdf_delta, 1e-12) is never actually used as output, it's purely there to get around the tf.select() gradient issue
    # if the probability on a sub-pixel is below 1e-5, we use an approximation based on the assumption that the log-density is constant in the bin of the observed sub-pixel value
    log_probs = tf.where(x < -0.999, log_cdf_plus, tf.where(x > 0.999, log_one_minus_cdf_min, tf.where(cdf_delta > 1e-5, tf.log(tf.maximum(cdf_delta, 1e-12)), log_pdf_mid - np.log(itg_interval/2.))))

    log_probs = tf.reduce_sum(log_probs,3) + log_prob_from_logits(logit_probs)
    if sum_all:
        if return_prob:
            print(log_sum_exp(log_probs).get_shape().as_list())
            return -tf.reduce_sum(log_sum_exp(log_probs)), log_sum_exp(log_probs)
        else:
            return -tf.reduce_sum(log_sum_exp(log_probs))
    else:
        return -tf.reduce_sum(log_sum_exp(log_probs),[1,2])

def sample_from_discretized_mix_logistic_2(l,nr_mix):
    ls = int_shape(l)
    xs = ls[:-1] + [2]
    # unpack parameters
    logit_probs = l[:, :, :, :nr_mix]
    l = tf.reshape(l[:,:,:,nr_mix:], xs[:-1] + [5,nr_mix])
    # sample mixture indicator from softmax
    sel = tf.one_hot(tf.argmax(logit_probs - tf.log(-tf.log(tf.random_uniform(logit_probs.get_shape(), minval=1e-5, maxval=1. - 1e-5))), 3), depth=nr_mix, dtype=tf.float32)
    sel = tf.reshape(sel, xs[:-1] + [1,nr_mix])
    # select logistic parameters
    means = tf.reduce_sum(l[:,:,:,:2,:]*sel,4)
    log_scales = tf.maximum(tf.reduce_sum(l[:,:,:,2:4,:]*sel,4), -7.)
    coeffs = tf.reduce_sum(tf.nn.tanh(l[:,:,:,4:,:])*sel,4)
    # sample from logistic & clip to interval
    # we don't actually round to the nearest 8bit value when sampling
    u = tf.random_uniform(means.get_shape(), minval=1e-5, maxval=1. - 1e-5)
    x = means + tf.exp(log_scales)*(tf.log(u) - tf.log(1. - u))
    x0 = tf.minimum(tf.maximum(x[:,:,:,0], -1.), 1.)
    x1 = tf.minimum(tf.maximum(x[:,:,:,1] + coeffs[:,:,:,0]*x0, -1.), 1.)
    #x2 = tf.minimum(tf.maximum(x[:,:,:,2] + coeffs[:,:,:,1]*x0 + coeffs[:,:,:,2]*x1, -1.), 1.)
    return tf.concat([tf.reshape(x0,xs[:-1]+[1]), tf.reshape(x1,xs[:-1]+[1])],3)


def discretized_mix_logistic_loss_1(x,l, itg_interval=255.0, sum_all=True,return_prob=False):
    """ log-likelihood for mixture of discretized logistics, assumes the data has been rescaled to [-1,1] interval """
    xs = int_shape(x) # true image
    ls = int_shape(l) # predicted distribution
    nr_mix = int(ls[-1] / 3) # here and below: unpacking the params of the mixture of logistics
    logit_probs = l[:,:,:,:nr_mix]
    
    l = tf.reshape(l[:,:,:,nr_mix:], xs[:-1] + [2,nr_mix])
    means = l[:,:,:,:1,:]
    log_scales = tf.maximum(l[:,:,:,1:2,:], -7.)
    
    x = tf.reshape(x, xs + [1]) + tf.zeros(xs + [nr_mix]) # here and below: getting the means and adjusting them based on preceding sub-pixels    
    
    means = tf.reshape(means[:,:,:,0,:], [xs[0],xs[1],xs[2],1,nr_mix])
    centered_x = x - means
    inv_stdv = tf.exp(-log_scales)
    plus_in = inv_stdv * (centered_x + 1./itg_interval)
    cdf_plus = tf.nn.sigmoid(plus_in)
    min_in = inv_stdv * (centered_x - 1./itg_interval)
    cdf_min = tf.nn.sigmoid(min_in)
    log_cdf_plus = plus_in - tf.nn.softplus(plus_in) # log probability for edge case of 0 (before scaling)
    log_one_minus_cdf_min = -tf.nn.softplus(min_in) # log probability for edge case of 255 (before scaling)
    cdf_delta = cdf_plus - cdf_min # probability for all other cases
    mid_in = inv_stdv * centered_x
    log_pdf_mid = mid_in - log_scales - 2.*tf.nn.softplus(mid_in) # log probability in the center of the bin, to be used in extreme cases (not actually used in our code)

    log_probs = tf.where(x < -0.999, log_cdf_plus, tf.where(x > 0.999, log_one_minus_cdf_min, tf.where(cdf_delta > 1e-5, tf.log(tf.maximum(cdf_delta, 1e-12)), log_pdf_mid - np.log(itg_interval/2.))))

    log_probs = tf.reduce_sum(log_probs,3) + log_prob_from_logits(logit_probs)
    if sum_all:
        if return_prob:
            print(log_sum_exp(log_probs).get_shape().as_list())
            return -tf.reduce_sum(log_sum_exp(log_probs)), log_sum_exp(log_probs)
        else:
            return -tf.reduce_sum(log_sum_exp(log_probs))
    else:
        return -tf.reduce_sum(log_sum_exp(log_probs),[1,2])


def sample_from_discretized_mix_logistic_1(l,nr_mix):
    ls = int_shape(l)
    xs = ls[:-1] + [1]
    # unpack parameters
    logit_probs = l[:, :, :, :nr_mix] # pi
    l = tf.reshape(l[:,:,:,nr_mix:], xs[:-1] + [2,nr_mix])

    # sample mixture indicator from softmax
    sel = tf.one_hot(tf.argmax(logit_probs - tf.log(-tf.log(tf.random_uniform(logit_probs.get_shape(), minval=1e-5, maxval=1. - 1e-5))), 3), depth=nr_mix, dtype=tf.float32)
    sel = tf.reshape(sel, xs[:-1] + [1,nr_mix])

    # select logistic parameters
    means = tf.reduce_sum(l[:,:,:,:1,:]*sel,4)
    log_scales = tf.maximum(tf.reduce_sum(l[:,:,:,1:,:]*sel,4), -7.)
    
    # we don't actually round to the nearest 8bit value when sampling
    u = tf.random_uniform(means.get_shape(), minval=1e-5, maxval=1. - 1e-5)
    x = means + tf.exp(log_scales)*(tf.log(u) - tf.log(1. - u))

    x0 = tf.minimum(tf.maximum(x[:,:,:,0], -1.), 1.)
    
    return tf.reshape(x0, xs)

def discretized_mix_logistic_loss_idp_2(x, l, itg_interval=255.0, sum_all=True, return_prob=False):
    """
    1. this loss assume that real part and the imaginary part are independent of each other
    """
    
    xs = int_shape(x)
    ls = int_shape(l)

    # unpacking the params of the mixture of logistics
    nr_mix = int(ls[-1] / 5) 
    logit_probs = l[:,:,:,:nr_mix]
    l = tf.reshape(l[:,:,:,nr_mix:], xs[:-1] + [4,nr_mix])
    means = l[:,:,:,:2,:]
    log_scales = tf.maximum(l[:,:,:,2:4,:], -7.)

    # adjusting means
    x = tf.reshape(x, xs + [1]) + tf.zeros(xs + [nr_mix]) 
    centered_x = x - means
    inv_stdv = tf.exp(-log_scales)
    plus_in = inv_stdv * (centered_x + 1./itg_interval)
    cdf_plus = tf.nn.sigmoid(plus_in)
    min_in = inv_stdv * (centered_x - 1./itg_interval)
    cdf_min = tf.nn.sigmoid(min_in)
    log_cdf_plus = plus_in - tf.nn.softplus(plus_in) # log probability for edge case of 0 (before scaling)
    log_one_minus_cdf_min = -tf.nn.softplus(min_in) # log probability for edge case of 255 (before scaling)
    cdf_delta = cdf_plus - cdf_min # probability for all other cases
    mid_in = inv_stdv * centered_x
    log_pdf_mid = mid_in - log_scales - 2.*tf.nn.softplus(mid_in) # log probability in the center of the bin, to be used in extreme cases (not actually used in our code)

    # now select the right output: left edge case, right edge case, normal case, extremely low prob case (doesn't actually happen for us)

    # this is what we are really doing, but using the robust version below for extreme cases in other applications and to avoid NaN issue with tf.select()
    # log_probs = tf.select(x < -0.999, log_cdf_plus, tf.select(x > 0.999, log_one_minus_cdf_min, tf.log(cdf_delta)))

    # robust version, that still works if probabilities are below 1e-5 (which never happens in our code)
    # tensorflow backpropagates through tf.select() by multiplying with zero instead of selecting: this requires use to use some ugly tricks to avoid potential NaNs
    # the 1e-12 in tf.maximum(cdf_delta, 1e-12) is never actually used as output, it's purely there to get around the tf.select() gradient issue
    # if the probability on a sub-pixel is below 1e-5, we use an approximation based on the assumption that the log-density is constant in the bin of the observed sub-pixel value
    log_probs = tf.where(x < -0.999, log_cdf_plus, tf.where(x > 0.999, log_one_minus_cdf_min, tf.where(cdf_delta > 1e-5, tf.log(tf.maximum(cdf_delta, 1e-12)), log_pdf_mid - np.log(itg_interval/2.))))

    log_probs = tf.reduce_sum(log_probs,3) + log_prob_from_logits(logit_probs)
    if sum_all:
        if return_prob:
            print(log_sum_exp(log_probs).get_shape().as_list())
            return -tf.reduce_sum(log_sum_exp(log_probs)), log_sum_exp(log_probs)
        else:
            return -tf.reduce_sum(log_sum_exp(log_probs))
    else:
        return -tf.reduce_sum(log_sum_exp(log_probs),[1,2])


def discretized_mix_logistic_loss_bivariate(x, l, itg_interval=255.0, sum_all=True, return_prob=False):
    """log-likelihood for mixture of discretized bivariate logistics"""
    return None



from spreco.common.options import MODELS
from spreco.common import utils
from spreco.common.logger import logger


import tensorflow.compat.v1 as tf
tf.disable_eager_execution()

import os
import numpy as np
import time

class trainer():
    """
    """

    def __init__(self, train_pipe, test_pipe, config):

        self.train_pipe = train_pipe
        self.test_pipe  = test_pipe
        self.config     = config

        if train_pipe is not None:
            self.log_path   = utils.create_folder(config['log_folder'])
            utils.save_config(config, self.log_path)
            self.logger = logger(self.log_path)

        self.model      = None
        self.sess       = None
        self.pred_y     = None

        self.global_step = 0

    
    def get_model(self, export):
        """
        TODO: put this function into options
        """
        
        if self.config['model'] == MODELS.NCSN:
            tf.random.set_random_seed(self.config['seed'])
            from spreco.model.ncsn import ncsn as selected_class

        elif self.config['model'] == MODELS.SDE:
            from spreco.model.sde import sde as selected_class
        
        elif self.config['model'] == MODELS.PIXELCNN:
            from spreco.model.pixelcnn import pixelcnn as selected_class

        else:
            raise Exception("Currently, this model is not implemented!")

        self.model = selected_class(self.config)
        self.model.prep(export)

    def mk_feed_dict(self, pipe, reps=None):
        """"utility to make feed dict"""
        if reps is None:
            reps = self.config['nr_gpu']

        keys = self.model.ins_outs.keys()
        elm = pipe.get_next()
        feed_dict ={}

        for key in keys:
            tmp = np.split(elm[key], self.config['nr_gpu'])
            feed_dict.update({self.model.ins_outs[key][i]: tmp[i] for i in range(reps)})

        return feed_dict

    def train_loop(self):

        # ready to go
        init_op    = tf.global_variables_initializer()
        saver      = tf.train.Saver(max_to_keep=self.config['max_keep'])
        gpu_config = tf.ConfigProto(allow_soft_placement=True)
        sess       = tf.Session(config=gpu_config)
        sess.run(init_op)

        utils.print_parameters(self.log_path+'/layer_info')
        train_loss = []
        test_loss  = []
        begin      = time.time()

        while self.train_pipe.get_epoch() < self.config['max_epochs']:

            feed_dict = self.mk_feed_dict(self.train_pipe)

            learning_rate = utils.get_lr(self.global_step, self.config['lr'],
                                         warmup_steps=None if 'warmup_steps' not in self.config.keys() else self.config['warmup_steps'],
                                         hidden_size=None if 'hidden_size' not in self.config.keys() else self.config['hidden_size'])

            feed_dict.update({self.model.learning_rate: learning_rate})

            l, _      = sess.run([self.model.loss_train, self.model.train_op], feed_dict=feed_dict)
            self.global_step = self.global_step + 1

            kvs = {}
            kvs['train_loss'] = l
            kvs['learning_rate'] = learning_rate
            self.logger.writekvs(kvs)

            train_loss.append(l)

            if self.train_pipe.check_epoch():

                # one epoch is finised
                self.train_pipe.update_epoch()
                if self.train_pipe.get_epoch() % self.config["save_interval"] == 0:
                    saver.save(sess, os.path.join(self.log_path, self.config['saved_name']+'_'+str(self.train_pipe.get_epoch())))

                info = "Epochs %d, time = %ds, train loss = %.4f" % (self.train_pipe.get_epoch(), time.time()-begin, np.mean(train_loss))
                if self.config['print_loss']:
                    print(info)
                utils.log_to(os.path.join(self.log_path, 'loss'), [info], prefix="->")

                begin      = time.time()
                train_loss = []

                # run test
                if self.test_pipe is not None:
                    
                    while not self.test_pipe.check_epoch():
                        feed_dict = self.mk_feed_dict(self.test_pipe)
                        l         = sess.run(self.model.loss_test, feed_dict=feed_dict)

                        kvs = {}
                        kvs['test_loss'] = l
                        self.logger.writekvs(kvs)

                        test_loss.append(l)
                        
                    self.test_pipe.update_epoch()

                    info = "Epochs %d, time = %ds, test loss = %.4f" % (self.test_pipe.get_epoch(), time.time()-begin, np.mean(test_loss))
                    if self.config['print_loss']:
                        print(info)
                    else:
                        utils.log_to(os.path.join(self.log_path, 'loss'), [info], prefix="->")

                    test_loss  = []
                    begin      = time.time()

    def train(self):

        # set gpu
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = self.config['gpu_id']

        self.train_pipe.start()
        if self.test_pipe is not None:
            self.test_pipe.start()
        self.get_model(export=False)
        self.train_loop()
        self.train_pipe.stop()
        if self.test_pipe is not None:
            self.test_pipe.stop()

    def restore(self, model_path):
        """
        TODO: how to reuse the codes in function prep
        """
        print("Restoring model")
        saver = tf.train.Saver()
        gpu_options = tf.GPUOptions(allow_growth=True, visible_device_list='0')
        config_proto = tf.ConfigProto(gpu_options=gpu_options)
        config_proto.gpu_options.allow_growth=True
        serialized = config_proto.SerializeToString()
        gpu_id = list(map(hex, serialized))
        sess = tf.Session(config=config_proto)
        sess.run(tf.global_variables_initializer())
        saver.restore(sess, model_path)
        return saver, sess, gpu_id

    def export(self, model_path, export_path, name):
        
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = self.config['gpu_id']

        self.get_model(export=True)
        saver, sess, gpu_id = self.restore(model_path)
        utils.export_model(saver, sess, export_path, name, gpu_id=gpu_id)


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import os.path as osp
import time
import numpy as np
import tensorflow.compat.v1 as tf
from tensorflow.core.util import event_pb2

class logger(object):
    """
    Dumps key/value pairs into TensorBoard's numeric format.

    Credit: OpenAI baselines. Licensed under MIT License.
    Copyright (c) 2017 OpenAI (http://openai.com)
    """

    def __init__(self, dir):
        os.makedirs(dir, exist_ok=True)
        self.dir = dir
        self.step = 1
        prefix = "events"
        path = osp.join(osp.abspath(dir), prefix)
        import tensorflow.compat.v1 as tf
        try:
            from tensorflow.python import _pywrap_events_writer as event_writer # v2
        except:
            print("using tensorflow 1.x")
            from tensorflow.python import pywrap_tensorflow as event_writer# v1
        else:
            print("using tensorflow 2.x")

        from tensorflow.core.util import event_pb2
        from tensorflow.python.util import compat

        self.tf = tf
        self.event_pb2 = event_pb2
        self.event_writer = event_writer
        self.writer = event_writer.EventsWriter(compat.as_bytes(path))

    def writekvs(self, kvs):
        def summary_val(k, v):
            kwargs = {"tag": k, "simple_value": float(v)}
            return self.tf.Summary.Value(**kwargs)

        summary = self.tf.Summary(value=[summary_val(k, v) for k, v in kvs.items()])
        event = self.event_pb2.Event(wall_time=time.time(), summary=summary)
        event.step = self.step
        self.writer.WriteEvent(event)
        self.writer.Flush()
        self.step += 1

    def close(self):
        if self.writer:
            self.writer.Close()
            self.writer = None

    @staticmethod
    def make_image(tensor, scalar):
        """Convert an numpy representation image to Image protobuf"""
        from PIL import Image
        if len(tensor.shape) == 4:
            _, height, width, channel = tensor.shape
        elif len(tensor.shape) == 3:
            height, width, channel = tensor.shape
        elif len(tensor.shape) == 2:
            height, width = tensor.shape
            channel = 1
        tensor = (tensor*scalar).astype(np.uint8)
        image = Image.fromarray(tensor)
        import io
        output = io.BytesIO()
        image.save(output, format='PNG')
        image_string = output.getvalue()
        output.close()
        return tf.Summary.Image(height=height, width=width, colorspace=channel, encoded_image_string=image_string)

    def log_image(self, im, tag, scalar=1, step=0):
        im = self.make_image(im, scalar)

        summary = [tf.Summary.Value(tag=tag, image=im)]
        summary = tf.Summary(value=summary)
        event = event_pb2.Event(summary=summary)
        event.step = step
        self.writer.WriteEvent(event)
        self.writer.Flush()


from setuptools import setup, find_packages

setup(
    name='spreco',
    version='0.0.1',
    description='Training priors for MRI image reconstruction',
    packages=find_packages(),
    install_requires=[
        'tensorflow',
        'tf-slim==1.1.0',
        'numpy',
        'pillow==8.2.0',
        'matplotlib',
        'scikit-image',
        'pyyaml==5.4.1',
        'tqdm'
    ],
)


from . import utils
import numpy as np

from multiprocessing import Queue, current_process, Process

class data_pipe(object):
    """
    parts may includes scale_func, noise_func, slice_func, normalize_func, file_loader, transform_func
    TODO: to make pipe compatible with the parts list that has two parallel streams
    """
    def __init__(self, parts, files, buffer_size, batch_size, shape_info, names=None):

        self.parts        = parts     # a list contains functionalities for data preparation (normalization, augmentation)
        self.files        = files     # a list contains paths of training files
        self.queue        = Queue(buffer_size)
        self.names        = names     # names of streams [inputs, labels]
        self.rseed        = None
        self.batch_size   = batch_size
        self.shape_info   = shape_info
        self.epoch_hook   = Queue(1)
        self.epoch        = 0
        self.p_steams     = len(parts) # number of streams of preprocess function
        self.nr_files     = len(files)
        self.idx          = 0
        self.is_next_file = True
        self.local_random = np.random.RandomState(self.rseed)
        self.data_idx     = self.local_random.permutation(self.nr_files)
        self.lefted       = [np.zeros((None)) for _ in range(self.p_steams)]
        self.batch        = [np.zeros(self.shape_info[i]) for i in range(self.p_steams)]
        self.cur_remained = 0

    def get_epoch(self):
        return self.epoch
    
    def update_epoch(self):
        self.epoch = self.epoch_hook.get()

    def check_epoch(self):
        return self.epoch_hook.full()

    def fetch_single(self, idx):
        pieces = [np.zeros((None)) for _ in range(self.p_steams)]
        for i in range(self.p_steams):
            parts = self.parts[i]
            flow  = self.files[self.data_idx[idx]]
            for part in parts:
                flow = part(flow)
            pieces[i] = flow
        return pieces

    def kernel(self):
        while True:
            while not self.queue.full():

                if self.is_next_file:

                    if self.idx == self.nr_files:
                        self.idx = 0
                        self.epoch = self.epoch + 1
                        self.epoch_hook.put(self.epoch)
                        self.data_idx = self.local_random.permutation(self.nr_files)
                        utils.color_print("Files permuted", 'red')
                    #print(self.idx, self.files[self.data_idx[self.idx]], current_process().name)
                    
                    pieces = [np.zeros((None)) for _ in range(self.p_steams)]
                    for i in range(self.p_steams):
                        parts = self.parts[i]
                        flow  = self.files[self.data_idx[self.idx]]
                        for part in parts:
                            flow = part(flow)
                        pieces[i] = flow
                    self.idx = self.idx + 1
                    

                    for i in range(self.p_steams):
                        if self.cur_remained == 0:
                            self.lefted = pieces  
                        else:
                            self.lefted[i] = np.concatenate([self.lefted[i], pieces[i]], axis=0)
                    self.cur_remained = self.check_remained(self.lefted)

                    if self.cur_remained > self.batch_size:
                        self.is_next_file = False
                        for i in range(self.p_steams):
                        
                            self.batch[i]  = self.lefted[i][0:self.batch_size, ...]
                            self.lefted[i] = self.lefted[i][self.batch_size:, ...]
                        
                        if self.names is not None:
                            tmp = {}
                            for i in range(self.p_steams):
                                tmp[self.names[i]] = self.batch[i]
                            self.queue.put(tmp)
                        else:
                            self.queue.put(self.batch)
                            
                        self.cur_remained = self.check_remained(self.lefted)
                        if self.cur_remained <= self.batch_size:
                            self.is_next_file = True

    def check_remained(self, lefted, idx=0):
        """
        return the number of batches remained in the list "lefted"
        """
        if lefted[idx].shape == ():
            cur_remained = 0
        else:
            cur_remained = lefted[idx].shape[0]
        return cur_remained

    def get_next(self):
        """
        return one batch which is a numpy array
        """
        return self.queue.get()
    
    def start(self):
        self.proc = Process(target=self.kernel)
        self.proc.daemon = True
        self.proc.start()
    
    def stop(self):
        self.proc.terminate()

def create_pipe(parts_funcs, files, batch_size, shape_info, buffer_size=10, names=None):
    pipe = data_pipe(parts=parts_funcs,
                       files=files,
                       buffer_size=buffer_size,
                       batch_size=batch_size,
                       shape_info=shape_info,
                       names=names)
    try:
        _ = pipe.fetch_single(0)
        return pipe
    except:
        raise Exception("Creating pipe failed, please check your parts funcs")


from spreco.model.utils import *

import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tf_slim import add_arg_scope
import numpy as np


@add_arg_scope
def dense(x_, num_units, nonlinearity=None, init_scale=1., counters={}, init=False, use_bias=True, ema=None, **kwargs):
    ''' fully connected layer '''
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('dense', counters)

    if 'debug' in kwargs.keys():
        if kwargs['debug']:
            print(name)
    stop_grad = False 
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    with tf.variable_scope(name):
        V = get_variable('V', stop_grad, shape=[int(x_.get_shape()[1]),num_units], dtype=tf.float32,
                              initializer=tf.random_normal_initializer(0, 0.05), trainable=True)
        g = get_variable('g', stop_grad, shape=[num_units], dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
        if use_bias:
            b = get_variable('b', stop_grad, shape=[num_units], dtype=tf.float32,
                                initializer=tf.constant_initializer(0.), trainable=True)

        # use weight normalization (Salimans & Kingma, 2016)
        # https://arxiv.org/pdf/1602.07868.pdf
        x = tf.matmul(x_, V)

        scaler = g / tf.sqrt(tf.reduce_sum(tf.square(V), [0]))

        if use_bias:
            x = tf.reshape(scaler, [1, num_units]) * x + tf.reshape(b, [1, num_units])

        if init: # normalize x
            m_init, v_init = tf.nn.moments(x, [0])
            scale_init = init_scale/tf.sqrt(v_init + 1e-10)
            if use_bias:
                with tf.control_dependencies([g.assign(g*scale_init), b.assign_add(-m_init*scale_init)]):
                    x = tf.matmul(x_, V)
                    scaler = g / tf.sqrt(tf.reduce_sum(tf.square(V), [0]))
                    x = tf.reshape(scaler, [1, num_units]) * x + tf.reshape(b, [1, num_units])
            else:
                with tf.control_dependencies([g.assign(g*scale_init)]):
                    x = tf.matmul(x_, V)
                    scaler = g / tf.sqrt(tf.reduce_sum(tf.square(V), [0]))
                    x = tf.reshape(scaler, [1, num_units]) * x

        # apply nonlinearity
        if nonlinearity is not None:
            x = nonlinearity(x)

        return x

@add_arg_scope
def dense_plus(x_, h, num_units, nr_classes, nonlinearity=None, init_scale=1., counters={}, init=False, use_bias=True, ema=None, **kwargs):
    ''' fully connected layer '''
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('dense', counters)

    if 'debug' in kwargs.keys():
        if kwargs['debug']:
            print(name)
    stop_grad = False 
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    with tf.variable_scope(name):
        V = get_variable('V', stop_grad, shape=[nr_classes, int(x_.get_shape()[1]),num_units], dtype=tf.float32,
                              initializer=tf.random_normal_initializer(0, 0.05), trainable=True)
        g = get_variable('g', stop_grad, shape=[nr_classes, num_units], dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
        if use_bias:
            b = get_variable('b', stop_grad, shape=[nr_classes, num_units], dtype=tf.float32,
                                initializer=tf.constant_initializer(0.), trainable=True)

        V_h = tf.gather(V, h, axis=0)
        g_h = tf.gather(g, h, axis=0)
        if use_bias:
            b_h =tf.gather(b, h, axis=0)

        # use weight normalization (Salimans & Kingma, 2016)
        x = tf.einsum('ij,ijk->ik',x_, V_h)

        scaler = g_h / tf.sqrt(tf.reduce_sum(tf.square(V_h), [1]))

        if use_bias:
            x = scaler * x + b_h

        # apply nonlinearity
        if nonlinearity is not None:
            x = nonlinearity(x)

        return x


@add_arg_scope
def conv2d_plus(x_, num_filters, filter_size=[3,3], stride=[1,1], pad='SAME', nonlinearity=None, init_scale=1., counters={}, init=False, ema=None, bias=True, weight_norm=False, **kwargs):
    ''' convolutional layer '''
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters) # this is the scope defined by args
    else:
        name = get_name('conv2d', counters) # this is default scope named with conv2d
    
    if 'debug' in kwargs.keys():
        if kwargs['debug']:
            print(name)
    stop_grad = False 
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    dilation = 1
    if 'dilation' in kwargs.keys():
        dilation = kwargs['dilation']

    with tf.variable_scope(name):
        V = get_variable('V', stop_grad, shape=filter_size+[int(x_.get_shape()[-1]),num_filters], dtype=tf.float32,
                              initializer=tf.random_normal_initializer(0, 0.05), trainable=True)

        if weight_norm:
            g = get_variable('g', stop_grad, shape=[num_filters], dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
            # use weight normalization (Salimans & Kingma, 2016)
            W = tf.reshape(g, [1, 1, 1, num_filters]) * tf.nn.l2_normalize(V, [0, 1, 2])
        else:
            W = V
        
        # calculate convolutional layer output
        x = tf.nn.conv2d(x_, W, [1] + stride + [1], pad, dilations=dilation)

        if bias:
            b = get_variable('b', stop_grad, shape=[num_filters], dtype=tf.float32,
                              initializer=tf.constant_initializer(0.), trainable=True)
            x = tf.nn.bias_add(x, b)

        if init and weight_norm:  # normalize x
            m_init, v_init = tf.nn.moments(x, [0,1,2])
            scale_init = init_scale / tf.sqrt(v_init + 1e-10)
            with tf.control_dependencies([g.assign(g * scale_init), b.assign_add(-m_init * scale_init)]):
                # x = tf.identity(x)
                W = tf.reshape(g, [1, 1, 1, num_filters]) * tf.nn.l2_normalize(V, [0, 1, 2])
                x = tf.nn.conv2d(x_, W, [1] + stride + [1], pad, dilations=dilation)
                if bias:
                    x = tf.nn.bias_add(x, b)

        # apply nonlinearity
        if nonlinearity is not None:
            x = nonlinearity(x)

        return x

@add_arg_scope
def conv2d(x_, num_filters, filter_size=[3,3], stride=[1,1], pad='SAME', nonlinearity=None, init_scale=1., counters={}, init=False, ema=None, bias=True, **kwargs):
    ''' convolutional layer '''
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters) # this is the scope defined by args
    else:
        name = get_name('conv2d', counters) # this is default scope named with conv2d
    
    if 'debug' in kwargs.keys():
        if kwargs['debug']:
            print(name)
    stop_grad = False 
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    dilation = 1
    if 'dilation' in kwargs.keys():
        dilation = kwargs['dilation']

    with tf.variable_scope(name):
        V = get_variable('V', stop_grad, shape=filter_size+[int(x_.get_shape()[-1]),num_filters], dtype=tf.float32,
                              initializer=tf.random_normal_initializer(0, 0.05), trainable=True)
        g = get_variable('g', stop_grad, shape=[num_filters], dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
        b = get_variable('b', stop_grad, shape=[num_filters], dtype=tf.float32,
                              initializer=tf.constant_initializer(0.), trainable=True)

        # use weight normalization (Salimans & Kingma, 2016)
        W = tf.reshape(g, [1, 1, 1, num_filters]) * tf.nn.l2_normalize(V, [0, 1, 2])
        
        # calculate convolutional layer output
        x = tf.nn.conv2d(x_, W, [1] + stride + [1], pad, dilations=dilation)
        if bias:
            x = tf.nn.bias_add(x, b)

        if init:  # normalize x
            m_init, v_init = tf.nn.moments(x, [0,1,2])
            scale_init = init_scale / tf.sqrt(v_init + 1e-10)
            with tf.control_dependencies([g.assign(g * scale_init), b.assign_add(-m_init * scale_init)]):
                # x = tf.identity(x)
                W = tf.reshape(g, [1, 1, 1, num_filters]) * tf.nn.l2_normalize(V, [0, 1, 2])
                x = tf.nn.conv2d(x_, W, [1] + stride + [1], pad, dilations=dilation)
                if bias:
                    x = tf.nn.bias_add(x, b)

        # apply nonlinearity
        if nonlinearity is not None:
            x = nonlinearity(x)

        return x

@add_arg_scope
def deconv2d(x_, num_filters, filter_size=[3,3], stride=[1,1], pad='SAME', nonlinearity=None, init_scale=1., counters={}, init=False, ema=None, bias = True, **kwargs):
    ''' transposed convolutional layer '''
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('deconv2d', counters)

    if 'debug' in kwargs.keys():
        if kwargs['debug']:
            print(name)

    stop_grad = False 
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    xs = int_shape(x_)
    if pad=='SAME':
        target_shape = [xs[0], xs[1]*stride[0], xs[2]*stride[1], num_filters]
    else:
        target_shape = [xs[0], xs[1]*stride[0] + filter_size[0]-1, xs[2]*stride[1] + filter_size[1]-1, num_filters]
    with tf.variable_scope(name):
        V = get_variable('V', stop_grad, shape=filter_size+[num_filters,int(x_.get_shape()[-1])], dtype=tf.float32,
                              initializer=tf.random_normal_initializer(0, 0.05), trainable=True)
        g = get_variable('g', stop_grad, shape=[num_filters], dtype=tf.float32,
                                initializer=tf.constant_initializer(1.), trainable=True)
        b = get_variable('b', stop_grad, shape=[num_filters], dtype=tf.float32,
                              initializer=tf.constant_initializer(0.), trainable=True)

        # use weight normalization (Salimans & Kingma, 2016)
        W = tf.reshape(g, [1, 1, num_filters, 1]) * tf.nn.l2_normalize(V, [0, 1, 3])

        # calculate convolutional layer output
        x = tf.nn.conv2d_transpose(x_, W, target_shape, [1] + stride + [1], padding=pad)
        if bias:
            x = tf.nn.bias_add(x, b)

        if init:  # normalize x
            m_init, v_init = tf.nn.moments(x, [0,1,2])
            scale_init = init_scale / tf.sqrt(v_init + 1e-10)
            with tf.control_dependencies([g.assign(g * scale_init), b.assign_add(-m_init * scale_init)]):
                # x = tf.identity(x)
                W = tf.reshape(g, [1, 1, num_filters, 1]) * tf.nn.l2_normalize(V, [0, 1, 3])
                x = tf.nn.conv2d_transpose(x_, W, target_shape, [1] + stride + [1], padding=pad)
                if bias:
                    x = tf.nn.bias_add(x, b)

        # apply nonlinearity
        if nonlinearity is not None:
            x = nonlinearity(x)

        return x


@add_arg_scope
def nin(x, num_units, use_bias=True, nonlinearity=None, **kwargs):
    """ a network in network layer (1x1 CONV) """
    s = int_shape(x)
    if None in s:
        x = tf.reshape(x, [-1,s[-1]])  # 03.09 11:18
    else:
        x = tf.reshape(x, [np.prod(s[:-1]),s[-1]])

    x = dense(x, num_units, use_bias=use_bias, **kwargs)

    if nonlinearity is not None:
        x = nonlinearity(x)

    if None in s:
        if len(s) == 2:
            shape = [-1] + [num_units]
        else: 
            shape = [-1] +  list(map(int, s[1:-1])) + [num_units]
        out = tf.reshape(x, shape)
    else:
        out = tf.reshape(x, s[:-1]+[num_units])
    return out

@add_arg_scope
def emb(t, embedding_size=256, scale=1.0, counters={}, **kwargs):
    """
    gaussian fourier embedding for t
    # https://www.cs.cmu.edu/~schneide/DougalRandomFeatures_UAI2015.pdf
    # https://people.eecs.berkeley.edu/~brecht/papers/07.rah.rec.nips.pdf
    """

    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('emb', counters)

    if 'debug' in kwargs.keys():
        if kwargs['debug']:
            print(name)

    with tf.variable_scope(name):

        W = get_variable('W', stop_grad=True, shape=[embedding_size], dtype=tf.float32,
                            initializer=tf.random_normal_initializer(stddev=scale), trainable=True)
        t_proj = t[:, None]*W[None, :] * 2 * 3.1415926
        
        return tf.concat([tf.math.sin(t_proj), tf.math.cos(t_proj)], axis=-1)

@add_arg_scope
def embed_t(t, embedding_size=256, scale=1.0, counters={}, **kwargs):
    """
    gaussian fourier embedding for t
    # https://www.cs.cmu.edu/~schneide/DougalRandomFeatures_UAI2015.pdf
    # https://people.eecs.berkeley.edu/~brecht/papers/07.rah.rec.nips.pdf
    """
    name = get_name('embed_t', counters)
    W = get_variable(name, stop_grad=True, shape=[embedding_size], dtype=tf.float32,
                            initializer=tf.random_normal_initializer(stddev=scale), trainable=True)
    t_proj = t[:, None]*W[None, :] * 2 * 3.1415926
    return tf.concat([tf.math.sin(t_proj), tf.math.cos(t_proj)], axis=-1)

@add_arg_scope
def self_attention(x, qk_chns, v_chns, **kwargs):
    """
    Non local neural networks
    https://arxiv.org/pdf/1711.07971.pdf
    """
    shape = int_shape(x)
    if None in shape:
        flatten_shape = (-1, shape[1]*shape[2], shape[3])
        out_shape = (-1, shape[1], shape[2], v_chns)
    else:
        flatten_shape = (shape[0], shape[1]*shape[2], shape[3])
        out_shape = (shape[0], shape[1], shape[2], v_chns)

    query_conv = tf.reshape(nin(x, qk_chns, nonlinearity=None, scope='global_attention'), flatten_shape)
    key_conv   = tf.reshape(nin(x, qk_chns, nonlinearity=None, scope='global_attention'), flatten_shape)
    value_conv = tf.reshape(nin(x, v_chns, nonlinearity=None, scope='global_attention'), flatten_shape)

    correlation = tf.einsum("bnf,bjf->bnj", query_conv, key_conv)
    attention_map = tf.nn.softmax(correlation, axis=-1)
    out = tf.einsum("bnf,bnj->bjf", value_conv, attention_map)
    out = nin(out, v_chns, scope='global_attention')

    if shape[-1] != v_chns:
        x = nin(x, v_chns, nonlinearity=None, scope='global_attention')
        
    return tf.reshape(out, out_shape) + x


## normalization modules
@add_arg_scope
def batch_normalization(x, is_training=True, counters={}, **kwargs):
    
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('batch_norm', counters)

    if 'debug' in kwargs.keys():
        if kwargs['debug']:
            print(name)

    with tf.variable_scope(name):
        return tf.layers.batch_normalization(x, is_training)

@add_arg_scope
def cond_instance_norm_plus(x, h, nr_classes, counters={}, **kwargs):
    """
    Adjusted conditional instance normalization arXiv:1907.05600

    y is the index of classes
    """
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('cond_instance_norm_plus', counters)

    stop_grad = False
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    in_shape = int_shape(x)

    with tf.variable_scope(name):
        gamma = get_variable(name+'_gamma', stop_grad, shape=[nr_classes, in_shape[-1]], dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
        beta = get_variable(name+'_beta', stop_grad, shape=[nr_classes, in_shape[-1]], dtype=tf.float32,
                                initializer=tf.constant_initializer(0.), trainable=True)
        alpha = get_variable(name+'_alpha', stop_grad, shape=[nr_classes, in_shape[-1]], dtype=tf.float32,
                                initializer=tf.constant_initializer(1.), trainable=True)
        mean, variance = tf.nn.moments(x, [1, 2], keepdims=True)
        nx = tf.nn.batch_normalization(x, mean, variance, offset=None, scale=None, variance_epsilon=1e-12, name='x_norm')
        cm, cvar = tf.nn.moments(mean, [-1], keepdims=True)
        adjusted_mean = tf.nn.batch_normalization(mean, cm, cvar, offset=None, scale=None, variance_epsilon=1e-12, name='adjust_norm')
        offset = tf.expand_dims(tf.expand_dims(tf.gather(beta, h, axis=0),1),1)
        scale = tf.expand_dims(tf.expand_dims(tf.gather(gamma, h, axis=0),1),1)
        out = nx + adjusted_mean*tf.expand_dims(tf.expand_dims(tf.gather(alpha, h, axis=0),1),1)
        #out = out*scale + offset
        out = tf.nn.batch_normalization(out, tf.zeros_like(mean), tf.ones_like(variance), offset=offset, scale=scale, variance_epsilon=1e-12, name='instancenorm') 
        
        return out

@add_arg_scope
def group_norm(x, groups=32, counters={}, **kwargs):

    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('group_norm', counters)
    
    in_shape = int_shape(x)

    stop_grad = False
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    channels = in_shape[-1]

    if groups > channels:
        raise ValueError('Invalid groups %d for %d channels' % (groups, channels))

    new_shape = in_shape[:-1] + [groups, channels//groups]
    x = tf.reshape(x, new_shape)

    params_shape = [1, 1, 1, groups, channels//groups]

    with tf.variable_scope(name):

        gamma = get_variable(name+'_gamma', stop_grad, shape=params_shape, dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
        beta = get_variable(name+'_beta', stop_grad, shape=params_shape, dtype=tf.float32,
                                initializer=tf.constant_initializer(0.), trainable=True)
        mean, variance = tf.nn.moments(x, [1,2,4], keepdims=True)

        x = tf.nn.batch_normalization(x, mean, variance, offset = beta, scale=gamma, variance_epsilon=1e-12, name='group_norm')
    
    out = tf.reshape(x, in_shape)

    return out

@add_arg_scope
def instance_norm_plus(x, counters={}, **kwargs):
    """
    Adjusted conditional instance normalization arXiv:1907.05600

    y is the index of classes
    """
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('instance_norm_plus', counters)

    stop_grad = False
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    in_shape = int_shape(x)

    with tf.variable_scope(name):
        gamma = get_variable(name+'_gamma', stop_grad, shape=[in_shape[-1]], dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
        beta = get_variable(name+'_beta', stop_grad, shape=[in_shape[-1]], dtype=tf.float32,
                                initializer=tf.constant_initializer(0.), trainable=True)
        alpha = get_variable(name+'_alpha', stop_grad, shape=[in_shape[-1]], dtype=tf.float32,
                                initializer=tf.constant_initializer(1.), trainable=True)
        mean, variance = tf.nn.moments(x, [1, 2], keepdims=True)
        nx = tf.nn.batch_normalization(x, mean, variance, offset=None, scale=None, variance_epsilon=1e-12, name='x_norm')
        cm, cvar = tf.nn.moments(mean, [-1], keepdims=True)
        adjusted_mean = tf.nn.batch_normalization(mean, cm, cvar, offset=None, scale=None, variance_epsilon=1e-12, name='adjust_norm')
        offset = tf.expand_dims(tf.expand_dims(tf.expand_dims(beta, 0),0),0)
        scale = tf.expand_dims(tf.expand_dims(tf.expand_dims(gamma, 0),0),0)
        out = nx + adjusted_mean*tf.expand_dims(tf.expand_dims(tf.expand_dims(alpha, 0), 0), 0)
        #out = out*scale + offset
        out = tf.nn.batch_normalization(out, tf.zeros_like(mean), tf.ones_like(variance), offset=offset, scale=scale, variance_epsilon=1e-12, name='instancenorm') 
        
        return out

@add_arg_scope
def instance_norm(x, counters={}, **kwargs):
    """
    Adjusted conditional instance normalization arXiv:1907.05600

    y is the index of classes
    """
    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('instance_norm_plus', counters)

    stop_grad = False
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    in_shape = int_shape(x)

    with tf.variable_scope(name):
        gamma = get_variable(name+'_gamma', stop_grad, shape=[in_shape[-1]], dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
        beta = get_variable(name+'_beta', stop_grad, shape=[in_shape[-1]], dtype=tf.float32,
                                initializer=tf.constant_initializer(0.), trainable=True)
        mean, variance = tf.nn.moments(x, [1, 2], keepdims=True)
        offset = tf.expand_dims(tf.expand_dims(tf.expand_dims(beta, 0),0),0)
        scale = tf.expand_dims(tf.expand_dims(tf.expand_dims(gamma, 0),0),0)
        out = tf.nn.batch_normalization(x, mean, variance, offset=offset, scale=scale, variance_epsilon=1e-12, name='instance_norm') 
        
        return out

@add_arg_scope
def layer_norm(x, counters={}, **kwargs):

    if 'scope' in kwargs.keys():
        name = get_name(kwargs['scope'], counters)
    else:
        name = get_name('layer_norm', counters)

    stop_grad = False
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    in_shape = int_shape(x)
    with tf.variable_scope(name):
        gamma = get_variable(name+'_gamma', stop_grad, shape=[in_shape[-1]], dtype=tf.float32,
                              initializer=tf.constant_initializer(1.), trainable=True)
        beta = get_variable(name+'_beta', stop_grad, shape=[in_shape[-1]], dtype=tf.float32,
                                initializer=tf.constant_initializer(0.), trainable=True)
        mean, variance = tf.nn.moments(x, [1,2,3], keepdims=True)
        out = tf.nn.batch_normalization(x, mean, variance, offset=beta, scale=gamma, variance_epsilon=1e-12, name='layer_norm')
        return out


##### PIXELCNN++ #####
@add_arg_scope
def gated_resnet(x, a=None, h=None, nonlinearity=concat_elu, conv=conv2d, init=False, counters={}, ema=None, dropout_p=0., **kwargs):
    xs = int_shape(x)
    num_filters = xs[-1]
    
    c1 = conv(nonlinearity(x), num_filters)
    if a is not None: # add short-cut connection if auxiliary input 'a' is given
        c1 += nin(nonlinearity(a), num_filters)
    c1 = nonlinearity(c1)
    if dropout_p > 0:
        c1 = tf.nn.dropout(c1, keep_prob=1. - dropout_p)
    
    c2 = conv(c1, num_filters * 2, init_scale=0.1)

    # add projection of h vector if included: conditional generation
    stop_grad = False 
    if 'stop_grad' in kwargs.keys():
        stop_grad = kwargs['stop_grad']

    if h is not None:
        with tf.variable_scope(get_name('conditional_weights', counters)):
            hw = get_variable('hw', stop_grad, ema, shape=[int_shape(h)[-1], 2 * num_filters], dtype=tf.float32,
                                    initializer=tf.random_normal_initializer(0, 0.05), trainable=True)
        if init:
            hw = hw.initialized_value()
        c2 += tf.reshape(tf.matmul(h, hw), [xs[0], 1, 1, 2 * num_filters])
    

    a, b = tf.split(c2, 2, 3)
    c3 = a * tf.nn.sigmoid(b)
    
    return x + c3


@add_arg_scope
def down_shifted_conv2d(x, num_filters, filter_size=[2,3], stride=[1,1], **kwargs):
    x = tf.pad(x, [[0,0],[filter_size[0]-1,0], [int((filter_size[1]-1)/2),int((filter_size[1]-1)/2)],[0,0]])
    return conv2d(x, num_filters, filter_size=filter_size, pad='VALID', stride=stride, **kwargs)

@add_arg_scope
def down_shifted_deconv2d(x, num_filters, filter_size=[2,3], stride=[1,1], **kwargs):
    x = deconv2d(x, num_filters, filter_size=filter_size, pad='VALID', stride=stride, **kwargs)
    xs = int_shape(x)
    return x[:,:(xs[1]-filter_size[0]+1),int((filter_size[1]-1)/2):(xs[2]-int((filter_size[1]-1)/2)),:]

@add_arg_scope
def down_right_shifted_conv2d(x, num_filters, filter_size=[2,2], stride=[1,1], **kwargs):
    x = tf.pad(x, [[0,0],[filter_size[0]-1, 0], [filter_size[1]-1, 0],[0,0]])
    return conv2d(x, num_filters, filter_size=filter_size, pad='VALID', stride=stride, **kwargs)

@add_arg_scope
def down_right_shifted_deconv2d(x, num_filters, filter_size=[2,2], stride=[1,1], **kwargs):
    x = deconv2d(x, num_filters, filter_size=filter_size, pad='VALID', stride=stride, **kwargs)
    xs = int_shape(x)
    return x[:,:(xs[1]-filter_size[0]+1):,:(xs[2]-filter_size[1]+1),:]
